/**
  *  \file client/dialogs/simalliances.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/simalliances.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/racelist.h"
#include "client/widgets/simalliancegrid.h"
#include "game/sim-if.h"
#include "game/sim.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/cardgroup.h"
#include "ui/widgets/cardtab.h"
#include "ui/widgets/checkbox.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/quit.h"
#include "ui/window.h"
#include "util/nls.h"

namespace {
    class WSimAllianceEditor {
     public:
        WSimAllianceEditor(GSimOptions& opts, bool& sync, GSimulatorGameInterface& game_if);

        void run();
        void onAllianceClick(int pl, int ally);
        void onEnemyClick(int pl, int ally);
        void onClick(int pl, int ally, GPlayerBitMatrix& mtx);
        void onAutoClick();
        void setAllianceData();
        void setEnemyData();
        UIGroup& makeGrid(ObjectHolder& h, int id, WSimAllianceGrid& grid, const string_t& text);

     private:
        GSimOptions& m_opts;
        bool& m_sync;
        GSimulatorGameInterface& m_interface;

        GPlayerBitMatrix m_alliances;
        GPlayerBitMatrix m_enemies;

        WSimAllianceGrid m_allianceGrid;
        WSimAllianceGrid m_enemyGrid;
        UICheckbox m_symmetricCheckbox;
        UICheckbox m_autoCheckbox;
    };
}

WSimAllianceEditor::WSimAllianceEditor(GSimOptions& opts, bool& sync, GSimulatorGameInterface& game_if)
    : m_opts(opts),
      m_sync(sync),
      m_interface(game_if),
      m_alliances(opts.alliance_settings),
      m_enemies(opts.enemy_settings),
      m_allianceGrid(),
      m_enemyGrid(),
      m_symmetricCheckbox(1001, _("Symmetric relations"), 's', 2),
      m_autoCheckbox(1002, _("Automatic"), 'a', 2)
{
    m_allianceGrid.sig_click.add(this, &WSimAllianceEditor::onAllianceClick);
    m_enemyGrid.sig_click.add(this, &WSimAllianceEditor::onEnemyClick);
    m_autoCheckbox.sig_changed.add(this, &WSimAllianceEditor::onAutoClick);
    m_symmetricCheckbox.setValue(1);
    m_autoCheckbox.setValue(sync);
}

void
WSimAllianceEditor::run()
{
    ObjectHolder h;

    // UIWindow [VBox]
    //   UICardGroupTab tab
    //   UICardGroup
    //     UIGroup[Grid] x 2
    //        UISpacer    WRaceList
    //        WRaceList   WSimAllianceGrid
    //   UICheckbox "symmetric"
    //   UICheckbox "auto"
    //   UIGroup[UIHBoxLayout]
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UISpacer
    UIWindow& window = h.add(new UIWindow(_("Edit Alliances"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    window.setLayout(UIVBoxLayout::instance5);

    string_t suffix = string_t("\n\n") + _("[PgUp]/[PgDn] to switch page.");
    UIGroup& g1a = makeGrid(h, 2000, m_allianceGrid, _("If a player (row) offers an alliance to another one (column), "
                                                       "their units will not attack the ally's units.") + suffix);
    UIGroup& g1e = makeGrid(h, 2001, m_enemyGrid,    _("If a player (row) declares another one (column) their enemy, "
                                                       "ships with Primary Enemy set will behave as if the enemy were an additional Primary Enemy.") + suffix);
    UICardGroup& g1 = h.add(new UICardGroup(1000));
    g1.add(g1a);
    g1.add(g1e);

    UICardGroupTab& tab = h.add(new UICardGroupTab(g1));
    tab.addTab(2000, _("Alliances"));
    tab.addTab(2001, _("Enemies"));
    tab.setKeys(UICardGroupTab::UsePageKeys);
    window.add(tab);
    window.add(g1);
    window.add(m_symmetricCheckbox);
    window.add(m_autoCheckbox);

    UIGroup& g4 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g4.add(h.add(new UIButton(1, _("OK"),     SDLK_RETURN)).addStop());
    g4.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g4.add(h.add(new UISpacer()));
    // g4.add(h.add(new UIButton(_("Help"), 'h')));
    window.add(g4);
    window.add(h.add(new WHelpWidget("pcc2:teams")));
    window.add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Tab)));
    window.add(h.add(new UIFocusIterator(1001, 1002, UIFocusIterator::fi_Vert)));
    window.add(h.add(new UIQuit(0)));

    window.pack();

    g1.setFocusedWidget();

    setAllianceData();
    setEnemyData();

    ui_root->centerWidget(window);
    if (ui_root->runModal(window)) {
        m_opts.alliance_settings = m_alliances;
        m_opts.enemy_settings = m_enemies;
        m_sync = m_autoCheckbox.getValue() != 0;
    }
}

void
WSimAllianceEditor::onAllianceClick(int pl, int ally)
{
    onClick(pl, ally, m_alliances);
    setAllianceData();
}

void
WSimAllianceEditor::onEnemyClick(int pl, int ally)
{
    onClick(pl, ally, m_enemies);
    setEnemyData();
}

void
WSimAllianceEditor::onClick(int pl, int ally, GPlayerBitMatrix& mtx)
{
    bool value = !mtx.get(pl, ally);
    mtx.set(pl, ally, value);
    if (m_symmetricCheckbox.getValue() != 0) {
        mtx.set(ally, pl, value);
    }
    m_autoCheckbox.setValue(0);
}

void
WSimAllianceEditor::onAutoClick()
{
    if (m_autoCheckbox.getValue() != 0) {
        m_interface.setDefaultRelations(m_alliances, m_enemies);
        setAllianceData();
        setEnemyData();
    }
}

void
WSimAllianceEditor::setAllianceData()
{
    m_allianceGrid.setContent(m_alliances, COLOR_GREENSCALE+7);
}

void
WSimAllianceEditor::setEnemyData()
{
    m_enemyGrid.setContent(m_enemies, COLOR_DARKRED);
}

UIGroup&
WSimAllianceEditor::makeGrid(ObjectHolder& h, int id, WSimAllianceGrid& grid, const string_t& text)
{
    UIGroup& g = h.add(new UIGroup(h.add(new UIGridLayout(3)), id));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new WRaceList(0, WRaceList::HorizontalLayout | WRaceList::ShowLetters   | WRaceList::NoColors, 1, NUM_OWNERS)));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new WRaceList(0, WRaceList::VerticalLayout   | WRaceList::ShowRaceNames | WRaceList::NoColors, 1, NUM_OWNERS)));
    g.add(grid);
    g.add(h.add(new UIMultilineStatic(FONT_NORMAL, text, font_ems[FONT_NORMAL] * 15, 0)));
    g.add(h.add(new UISpacer()));
    grid.setFocusedWidget();
    return g;
}

void
editAlliances(GSimOptions& opts, bool& sync, GSimulatorGameInterface& game_if)
{
    WSimAllianceEditor(opts, sync, game_if).run();
}
