/**
  *  \file client/dlg-structure.cc
  *  \brief Structure Building Dialog
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-structure.h"
#include "util/nls.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "cpluslib/holder.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "game/global.h"
#include "ui/msgbox.h"
#include "ui/skin.h"
#include "ui/util.h"
#include "util/misc.h"
#include "game/host.h"
#include "game/config.h"
#include "game/tables.h"
#include "cpluslib/format.h"
#include "client/dlg-tax.h"
#include "client/dlg-supplies.h"
#include "ui/panel.h"
#include "client/screens.h"
#include "game/spec.h"
#include "ui/prefix.h"
#include "client/dlg-groundcombat.h"
#include "game/actions/planetstructs.h"
#include "ui/widgets/focusiterator.h"
#include "resmgr/resmgr.h"
#include "client/tiles/planetmining.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "ui/widgets/cardgroup.h"
#include "client/widgets/helpwidget.h"
#include "client/dialogs/goaldialog.h"
#include "game/formula.h"
#include "game/planetform.h"
#include "game/planettransfer.h"
#include "util/strtab.h"
#include "ui/widgets/quit.h"

namespace {
    /** Planetary Mining Tile for Structure Building. This is like WPlanetMiningTile,
        but automatically updates the number of mines (see WPlanetMiningTile::setNumMines)
        from a GPlanetBuildStructuresAction. */
    class WPlanetStructureBuildMiningTile : public WPlanetMiningTile {
        GPlanetBuildStructuresAction& action;
        SignalConnection conn;
     public:
        WPlanetStructureBuildMiningTile(GPlanetBuildStructuresAction& action, GCargoType ele)
            : WPlanetMiningTile(action.getPlanet(), ele),
              action(action),
              conn(action.sig_changed.add(this, &WPlanetStructureBuildMiningTile::doUpdate))
            {
                doUpdate();
            }
        void doUpdate()
            {
                setNumMines(mp16_t(action.getAmount(str_Mines)));
            }
    };
}

/************************* WPlanetStructureWidget ************************/

WPlanetStructureWidget::WPlanetStructureWidget(GPlanetBuildStructuresAction& action,
                                               TPlanetaryStructure structure)
    : action(action), structure(structure)
{
    conn = observe(action.sig_changed);
}

WPlanetStructureWidget::~WPlanetStructureWidget()
{ }

void
WPlanetStructureWidget::add(int arg)
{
    action.addStructuresLimitCash(structure, arg ? arg : 1);
}

void
WPlanetStructureWidget::remove(int arg)
{
    action.addStructuresLimitCash(structure, arg ? -arg : -1);
}

void
WPlanetStructureWidget::drawContent(GfxCanvas& can)
{
    static const char*const costs[] = {
        // FIXME: I don't really like to hardcode these strings here
        N_("4 mc + 1 supply"),
        N_("3 mc + 1 supply"),
        N_("10 mc + 1 supply"),
        N_("10 mc + 1 Duranium")
    };
    const int x = getExtent().x, y = getExtent().y,
        y2 = y + font_heights[FONT_TITLE]+5, line = font_heights[FONT_NORMAL];

    const int goal = action.getPlanet().getAutobuildGoal(structure);
    const int max = getMaxStructures(action.getPlanet(), structure);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(UIColor::tc_Heading);
    outText(ctx, x, y, getPlanetaryStructureName(structure));

    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);
    outText(ctx, x, y2,        _("Amount:"));
    outText(ctx, x, y2+line,   _("Auto-B. Goal:"));
    outText(ctx, x, y2+2*line, _("Maximum:"));
    outText(ctx, x, y2+3*line, _("Cost:"));

    ctx.setColor(UIColor::tc_Green);
    outText(ctx, x+45, y2+3*line, _(costs[structure]));

    ctx.setSolidBackground();
    ctx.setTextAlign(2, 0);
    outTextF(ctx, x+150, y2,        40, numToString(action.getAmount(structure)));
    if (goal == MAX_AUTOBUILD_GOAL)
        outTextF(ctx, x+150, y2+line, 40, _("[max]"));
    else
        outTextF(ctx, x+150, y2+line, 40, numToString(goal));
    outTextF(ctx, x+150, y2+2*line, 40, numToString(max));

    can.update(getExtent());
}

bool
WPlanetStructureWidget::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && e.key.code == 'a') {
        action.doStandardAutoBuild();
        return true;
    }
    if (e.type == ev_Key && e.key.code == 'g') {
        GPlanet& planet = action.getPlanet();
        WBuildGoalsDialog window(false);
        window.initData(planet);
        window.setFocusToStructure(structure);
        ui_root->centerWidget(window);
        ui_root->add(window);
        if (ui_root->run()) {
            window.writeBack(planet);
            getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
        }
        return true;
    }
    if (e.type == ev_Key && e.key.code == 't') {
        doTaxation(action.getPlanet(),
                   action.getAmount(str_Mines),
                   action.getAmount(str_Factories));
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
        return true;
    }
    if (e.type == ev_Key && e.key.code == 's') {
        WSellSuppliesDialog::doDialog(action.getPlanet());
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
        return true;
    }
    if (e.type == ev_Key && e.key.code == 'c') {
        doGroundCombatPrediction(action.getPlanet(),
                                 action.getAmount(str_Defense));
        return true;
    }

    return UIWidget::handleEvent(e, second_pass);
}

void
WPlanetStructureWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.pref_size = info.min_size = GfxPoint(155,
                                              3*font_heights[FONT_NORMAL]
                                              + font_heights[FONT_TITLE] + 5);
}

/*********************** WPlanetStructureCostWidget **********************/

const char* const WPlanetStructureCostWidget::labels[] = {
    N_("You have:"),
    N_("Cost:"),
    N_("You will have\nremaining:"),
    N_("Duranium:\n  remaining:")
};
const char* const WPlanetStructureCostWidget::units[] = {
    N_("mc"),
    N_("kt supplies"),
    N_("kt")
};

WPlanetStructureCostWidget::WPlanetStructureCostWidget(GPlanetBuildStructuresAction& action,
                                                       bool with_duranium)
    : action(action),
      with_duranium(with_duranium)
{
    conn_action = observe(action.sig_changed);
    conn_planet = observe(action.getPlanet().sig_changed);
    computeLayout();
}

void
WPlanetStructureCostWidget::computeLayout()
{
    int n;

    // Label width
    label_width = 0;
    for (int i = 0; i < 4; ++i)
        if ((n = getTextWidthFormatted(*fonts[FONT_NORMAL], 99999, _(labels[i]))) > label_width)
            label_width = n;

    // Number width. Includes number plus spacing.
    number_width = font_ems[FONT_NORMAL] * 6;

    // Unit width
    unit_width = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(units, countof(units)));
}

WPlanetStructureCostWidget::~WPlanetStructureCostWidget()
{ }

void
WPlanetStructureCostWidget::drawContent(GfxCanvas& can)
{
    const int dy_line  = font_heights[FONT_NORMAL];
    const int dy_block = dy_line * 5/2;

    /*
        You have:      nn mc
                       nn kt supplies
        Cost:          nn mc
                       nn kt supplies
        You will have  nn mc
        remaining:     nn kt supplies
        Duranium:      nn kt
          remaining:   nn kt
    */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());

    // Left column
    int x = extent.x;
    int y = extent.y;
    ctx.setColor(UIColor::tc_Static);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextFormatted(ctx, x, y,              label_width, _(labels[0]));
    outTextFormatted(ctx, x, y + dy_block,   label_width, _(labels[1]));
    outTextFormatted(ctx, x, y + 2*dy_block, label_width, _(labels[2]));
    if (with_duranium)
        outTextFormatted(ctx, x, y + 3*dy_block, label_width, _(labels[3]));

    // Right column
    x = extent.x + label_width + number_width;
    ctx.setColor(UIColor::tc_Green);
    outText(ctx, x, y,                        _(units[0]));
    outText(ctx, x, y + dy_line,              _(units[1]));
    outText(ctx, x, y + dy_block,             _(units[0]));
    outText(ctx, x, y + dy_block + dy_line,   _(units[1]));
    outText(ctx, x, y + 2*dy_block,           _(units[0]));
    outText(ctx, x, y + 2*dy_block + dy_line, _(units[1]));
    if (with_duranium) {
        outText(ctx, x, y + 3*dy_block,           _(units[2]));
        outText(ctx, x, y + 3*dy_block + dy_line, _(units[2]));
    }

    // Middle column
    ctx.setTextAlign(2, 0);
    x -= fonts[FONT_NORMAL]->getTextWidth(" ");

    outText(ctx, x, y,                        numToString(action.getRemainingCargo(el_Money) + action.getReservedCargo(el_Money)));
    outText(ctx, x, y + dy_line,              numToString(action.getRemainingCargo(el_Supplies) + action.getReservedCargo(el_Supplies)));
    outText(ctx, x, y + dy_block,             numToString(action.getCost(el_Money)));
    outText(ctx, x, y + dy_block + dy_line,   numToString(action.getCost(el_Supplies)));
    outText(ctx, x, y + 2*dy_block,           numToString(action.getRemainingCargo(el_Money)));
    outText(ctx, x, y + 2*dy_block + dy_line, numToString(action.getRemainingCargo(el_Supplies)));
    if (with_duranium) {
        outText(ctx, x, y + 3*dy_block,           numToString(action.getRemainingCargo(el_Duranium) + action.getReservedCargo(el_Duranium)));
        outText(ctx, x, y + 3*dy_block + dy_line, numToString(action.getRemainingCargo(el_Duranium)));
    }

    can.update(getExtent());
}

void
WPlanetStructureCostWidget::getLayoutInfo(LayoutInfo& info)
{
    int lines       = with_duranium ? 4 : 3;
    int line_height = font_heights[FONT_NORMAL] * 5/2;

    info.pref_size = info.min_size = GfxPoint(label_width + number_width + unit_width,
                                              line_height * lines);
    info.flags = 0;
}

/********************** WPlanetStructureEffectWidget *********************/

WPlanetStructureEffectWidget::WPlanetStructureEffectWidget(GPlanetBuildStructuresAction& action)
    : action(action)
{
    conn_action = observe(action.sig_changed);
    conn_planet = observe(action.getPlanet().sig_changed);
}

WPlanetStructureEffectWidget::~WPlanetStructureEffectWidget()
{ }

void
WPlanetStructureEffectWidget::drawContent(GfxCanvas& can)
{
    /*
        Sensor visibility: n%, heavy
        Colonist tax:    n% (n mc)
        - They like your leadership
        Native tax:      n% (n mc)
        - They like your leadership
     */
    const GPlanet& planet = action.getPlanet();

    int mifa = action.getAmount(str_Mines) + action.getAmount(str_Factories);

    // Compute sensor visibility
    int dfu, mfd, ffd;
    if (host.isPHost()) {
        dfu = config.DefenseForUndetectable();
        mfd = config.MinesForDetectable();
        ffd = config.FactoriesForDetectable();
    } else {
        dfu = 15;
        mfd = 21;
        ffd = 16;
    }

    int rate;
    if (action.getAmount(str_Defense) >= dfu
        || dfu <= 0    /* probably not needed, but avoids possible divide-by-zero below. PCC 1.x has it, too. */
        || (action.getAmount(str_Mines) < mfd && action.getAmount(str_Factories) < ffd))
    {
        rate = 0;
    } else {
        rate = 100 - action.getAmount(str_Defense) * 100L / dfu;
    }

    int level;
    if (host.isPHost()) {
        level = mifa / 30;
        if (level > ind_Heavy)
            level = ind_Heavy;
    } else {
        level = (mifa >= 100 ? ind_Heavy : ind_Light);
    }

    // Display sensor visibility
    int x = extent.x;
    int y = extent.y;
    int w = extent.w;
    const int dy = font_heights[FONT_NORMAL];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();

    const char* svlabel = _("Sensor visibility:");
    const int svwidth = ctx.getFont().getTextWidth(svlabel);
    ctx.setColor(UIColor::tc_Static);
    outText(ctx, x, y, svlabel);
    ctx.setColor(rate != 0 ? UIColor::tc_Yellow : UIColor::tc_Green);
    outTextF(ctx, x + svwidth, y, w - svwidth,
             format(" %d%%, %s", rate, getIndustryLevelName(level)));

    // Colonist tax
    int32 rem_inc;
    int change = getColonistChange(planet, planet.getColonistTax(), mifa);
    y += dy*3/2;
    ctx.setColor(UIColor::tc_Static);
    outText(ctx, x, y, _("Colonist Tax:"));
    ctx.setColor(UIColor::tc_Green);
    outTextF(ctx, x + svwidth, y, w - svwidth,
             format(_("%d%% (%d mc)"), planet.getColonistTax(),
                    numToString(getColonistDueLimited(planet, planet.getColonistTax(), rem_inc))));
    y += dy;
    ctx.setColor(change < 0 ? UIColor::tc_Red : UIColor::tc_Green);
    outTextF(ctx, x, y, w,
             /* This format string means: display numerical 'change' only
                if it's nonzero */
             format("- %s%!d%!0{ (%1$+d)%}", getHappyChangeName(change), change));

    // Native tax
    if (planet.getNativeRace()) {
        y += dy*3/2;
        change = getNativeChange(planet, planet.getNativeTax(), mifa);
        ctx.setColor(UIColor::tc_Static);
        outText(ctx, x, y, _("Native Tax:"));
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x + svwidth, y, w - svwidth,
                 format(_("%d%% (%d mc)"), planet.getNativeTax(),
                        numToString(getNativeDueLimited(planet, planet.getNativeTax(), rem_inc))));
        y += dy;
        ctx.setColor(change < 0 ? UIColor::tc_Red : UIColor::tc_Green);
        outTextF(ctx, x, y, w,
                 format("- %s%!d%!0{ (%1$+d)%}", getHappyChangeName(change), change));
    }

    can.update(getExtent());
}

void
WPlanetStructureEffectWidget::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(20, 6);
    info.flags     = li_FlexibleH;
}

/*********************** WPlanetDefenseEffectWidget **********************/

static bool
canDoFighterPrediction()
{
    // We can do this for THost
    if (host.isTHost()) {
        return true;
    }

    // We can do this for PHost and classic combat if all FighterBeamExplosive etc. are constant.
    // For 101% precision, we'd also have to check that the options are not experience-modified,
    // but that'd take away a useful feature. Note that Shield/DamageScaling can be variable, as
    // they're taken from the victim, which would be us.
    if (host.isPHost() && !config.AllowAlternativeCombat()) {
        return config.FighterBeamExplosive.isAllTheSame()
            && config.FighterBeamKill.isAllTheSame();
    }

    return false;
}

static void
computeFighterEffects(int mass, int owner, int& shield, int& damage)
{
    if (host.isTHost()) {
        // THost formula
        shield = rdivaddw(80*2,      mass+1, 1);
        damage = rdivaddw(80*shield, mass+1, 1);
    } else {
        // PHost non-AC formula
        int boom = (config.ShieldDamageScaling(owner) * config.FighterBeamExplosive(1)
                    + config.ShieldKillScaling(owner) * config.FighterBeamKill(1));
        shield = radiv(boom, mass+1) + 1;
        damage = radiv(shield * config.CrewKillScaling(owner), mass+1) + 1;
    }
}

static int
computeNumStructures(int d, int k)
{
    /* Solve d = k*n*(n-1) for n. Taken from PCC 1.x.
       Actually, this is n=Round(Sqrt(d/k)), but in integer only and thus
       with "perfect" precision.

       Note that this does not work for k>4; it returns one too much.
       The reason is that the actual formula is not
             d = k*n*(n-1)
       but   d = k*n*(n-1) + 0.25*k,
       which is the inverse of
             n = rountToInt(sqrt(d/k))
       or    n = trunc(sqrt(d/k) + 0.5)

       With  n = sqrt(d/k) + 0.5, we get
       
             (n - 0.5)**2     = d/k
             n*n - n - 0.25   = d/k
           k*n*(n-1) - 0.25*k = d

       Fixing this would mean starting with sum=1+(d-1)/4.
       (I still don't know where the 1 offset comes from...) */
    int n = 1;
    int sum = 1;
    while (sum < d) {
        sum += 2*k*n;
        ++n;
    }
    if (sum > d)
        --n;
    return n;
}

WPlanetDefenseEffectWidget::WPlanetDefenseEffectWidget(GPlanetBuildStructuresAction& action)
    : action(action)
{
    conn_action = observe(action.sig_changed);
}

WPlanetDefenseEffectWidget::~WPlanetDefenseEffectWidget()
{ }

/** Draw line for table.
    \param dy    Relative position from start of widget
    \param name  Name of item (left column)
    \param next  Number of additional structures needed, -1 if maximum (right column)
    \param have  Number of structures we have
    \param max   Maximum allowed number of structures */
void
WPlanetDefenseEffectWidget::showLine(GfxContext& ctx, int dy, string_t name, int next, int have, int max)
{
    GfxPoint pt = getExtent().getTopLeft();
    int w  = getExtent().w;
    int w5 = w/5;

    ctx.setColor(UIColor::tc_Green);
    ctx.setTextAlign(0, 0);
    outTextF(ctx, pt.x, pt.y + dy, w-w5, name);
    ctx.setTextAlign(2, 0);
    if (next < 0) {
        outTextF(ctx, pt.x + w, pt.y + dy, w5, _("(max)"));
    } else {
        if (next + have > max)
            ctx.setColor(UIColor::tc_Yellow);
        outTextF(ctx, pt.x + w, pt.y + dy, w5, format("+%d", next));
    }
}

void
WPlanetDefenseEffectWidget::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);
    outText(ctx, extent.x, extent.y, _("You have:"));
    ctx.setTextAlign(2, 0);
    outText(ctx, extent.x + extent.w, extent.y, _("Next at:"));
    drawHLine(ctx, extent.x, extent.y + font_heights[FONT_NORMAL]-2, extent.x + extent.w-1);

    int y = font_heights[FONT_NORMAL] + 2;

    int defenseWant = action.getAmount(str_Defense);
    int sbdWant     = action.getAmount(str_BaseDefense);
    int totalWant   = defenseWant + sbdWant;

    int defenseMax = action.getMax(str_Defense);
    int sbdMax     = action.getMax(str_BaseDefense);
    int totalMax   = defenseMax + sbdMax;

    int n, max, next, add;

    /* Beam count */
    n = computeNumStructures(totalWant, 3);
    max = (host.isPHost() && config.AllowAlternativeCombat() ? 20 : 10);
    if (n >= max) {
        n    = max;
        next = -1;
    } else {
        next = n*(n+1)*3+1 - totalWant;
    }
    showLine(ctx, y, format(_("%d beam%!1{s%}"), n), next, totalWant, totalMax);
    y += font_heights[FONT_NORMAL];

    /* Beam type */
    if (n > 0) {
        n = computeNumStructures(defenseWant, 2);
        if (action.getPlanet().hasBase()
            && action.getPlanet().getBaseTechLevel(tl_Beam).isKnown()
            && action.getPlanet().getBaseTechLevel(tl_Beam) > n)
        {
            n = action.getPlanet().getBaseTechLevel(tl_Beam);
        }
        if (n > 0) {
            if (n >= 10) {
                n    = 10;
                next = -1;
            } else {
                next = n*(n+1)*2+1 - defenseWant;
            }
            showLine(ctx, y, "  " + getBeam(n).getName(), next, defenseWant, defenseMax);
            y += font_heights[FONT_NORMAL];
        }
    }

    /* Fighter */
    n = computeNumStructures(defenseWant, 1);
    next = n*(n+1)+1 - defenseWant;
    add = 0;
    if (action.getPlanet().hasBase())
        add = action.getPlanet().getCargo(el_Fighters);
    showLine(ctx, y, format(_("%d fighter%!1{s%}"), n + add), next, defenseWant, defenseMax);
    y += font_heights[FONT_NORMAL];

    /* Bays */
    if (action.getPlanet().hasBase() && host.isPHost()) {
        n += 5;
    }
    if (n >= MAX_BAY_LIMIT) {
        n = MAX_BAY_LIMIT;
        next = -1;
    }
    showLine(ctx, y, format(_("%d fighter bay%!1{s%}"), n), next, defenseWant, defenseMax);
    y += font_heights[FONT_NORMAL];

    /* Torpedoes */
    if (host.isPHost() && config.PlanetsHaveTubes()) {
        /* Launchers */
        n = computeNumStructures(totalWant, 4);
        if (n >= 20) {
            n    = 20;
            next = -1;
        } else {
            next = n*(n+1)*4+1 - totalWant;
        }
        showLine(ctx, y, format(_("%d torpedo launcher%!1{s%}"), n), next, totalWant, totalMax);
        y += font_heights[FONT_NORMAL];

        if (n > 0) {
            /* Type */
            int tech = computeNumStructures(defenseWant, 2);
            if (action.getPlanet().hasBase()
                && action.getPlanet().getBaseTechLevel(tl_Torpedo).isKnown()
                && action.getPlanet().getBaseTechLevel(tl_Torpedo) > tech)
            {
                tech = action.getPlanet().getBaseTechLevel(tl_Torpedo);
            }
            if (tech >= 10) {
                tech = 10;
                next = -1;
            } else {
                next = tech*(tech+1)*2+1 - defenseWant;
            }
            showLine(ctx, y, "  " + getTorp(tech).getName(), next, defenseWant, defenseMax);
            y += font_heights[FONT_NORMAL];

            /* Torpedoes */
            int32 total = 0;
            if (action.getPlanet().hasBase() && config.UseBaseTorpsInCombat(action.getPlanet().getOwner())) {
                for (int i = 1; i <= NUM_TORPS; ++i)
                    total += action.getPlanet().getCargo(getCargoTypeFromTorpType(i)) * getTorp(i).getTorpCost();
                if (getTorp(tech).getTorpCost() > 0)
                    total /= getTorp(tech).getTorpCost();
            }

            int16_t level = action.getPlanet().unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
            int ppt = config.PlanetaryTorpsPerTube(action.getPlanet().getOwner()) + getExperienceBonus(config.EModPlanetaryTorpsPerTube, level);
            total += ppt * n;
            if (ppt > 0)
                next = n*(n+1)*4+1 - totalWant;
            else
                next = -1;

            showLine(ctx, y, format(_("%d torpedo%!1{es%}"), total), next, totalWant, totalMax);
            y += font_heights[FONT_NORMAL];
        }
    }

    /* Fighter effects */
    if (canDoFighterPrediction()) {
        // Compute current status
        int shield;
        int damage;
        int owner = action.getPlanet().getOwner();
        computeFighterEffects(100 + totalWant, owner, shield, damage);

        // Try to find better amounts
        int shieldWant = 0;
        int damageWant = 0;
        int n = totalWant;
        if (shield <= 1) {
            shieldWant = -1;
        }
        if (damage <= 1) {
            damageWant = -1;
        }
        while (shieldWant == 0 || damageWant == 0) {
            int s, d;
            ++n;
            computeFighterEffects(100 + n, owner, s, d);
            if (shieldWant == 0 && s != shield) {
                shieldWant = n - totalWant;
            }
            if (damageWant == 0 && d != damage) {
                damageWant = n - totalWant;
            }
        }

        // Show it
        showLine(ctx, y, format(_("%d%% shield loss from enemy fighter"), shield), shieldWant, totalWant, totalMax);
        y += font_heights[FONT_NORMAL];
        showLine(ctx, y, format(_("%d%% damage from enemy fighter"), damage), damageWant, totalWant, totalMax);
        y += font_heights[FONT_NORMAL];
    }

    drawBackground(ctx, GfxRect(extent.x, extent.y + y, extent.w, extent.h - y));
    can.update(extent);
}

void
WPlanetDefenseEffectWidget::getLayoutInfo(LayoutInfo& info)
{
    // Your have/next at:
    // 1. Beam count
    // 2. Beam type
    // 3. Fighters
    // 4. Fighter bays
    // 5. Torp launchers
    // 6. Torp type
    // 7. Torps
    // 8. Shield loss
    // 9. Damage
    // ==> 10 lines
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(22, 10);
}

/************************* WPlanetStructureGroup *************************/

WPlanetStructureGroup::WPlanetStructureGroup(int id,
                                             GPlanetBuildStructuresAction& action,
                                             TPlanetaryStructure structure)
    : UIFocusGroup(UINullLayout::instance, id),
      psw(action, structure), add_btn(0, "+", '+'),
      sub_btn(0, "-", '-'),
      structure(structure)
{
    add(psw);
    add(add_btn);
    add(sub_btn);
    add_btn.sig_fired.add(&psw, &WPlanetStructureWidget::add);
    sub_btn.sig_fired.add(&psw, &WPlanetStructureWidget::remove);
}

WPlanetStructureGroup::~WPlanetStructureGroup()
{ }

void
WPlanetStructureGroup::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(315, 92+10);
}

void
WPlanetStructureGroup::onResize(GfxRect old_size, bool by_layout)
{
    UIFocusGroup::onResize(old_size, by_layout);

    GfxRect newpos = getExtent();
    psw.setExtent(GfxRect(newpos.x + 120, newpos.y + 5, 165, 92), by_layout);
    add_btn.setExtent(GfxRect(newpos.x + 285, newpos.y +  5, 25, 25), by_layout);
    sub_btn.setExtent(GfxRect(newpos.x + 285, newpos.y + 35, 25, 25), by_layout);
}

void
WPlanetStructureGroup::drawContent(GfxCanvas& can)
{
    static const char*const resid[] = {
        "planet.mine",
        "planet.factory",
        "planet.defense",
        "base.defense"
    };

    if (Ptr<GfxPixmap> pix = requestPixmap(ResId(resid[structure])))
        pix->blitSized(can, extent.x+5, extent.y+5, 105, 93);
    can.update(GfxRect(extent.x+5, extent.y+5, 105, 93));
    UIFocusGroup::drawContent(can);
}

/************************* WPlanetStructureHeader ************************/

WPlanetStructureHeader::WPlanetStructureHeader(GPlanet& planet, UICardGroup& group)
    : planet(planet),
      group(group),
      btn_left("<", SDLK_LEFT),
      btn_right(">", SDLK_RIGHT)
{
    conn = observe(group.sig_focus_changed);

    // Caller will pack() the dialog, thus calling onResize() on this
    // object, which will fill in the actual button positions.
    addChildWidget(btn_left, 0);
    addChildWidget(btn_right, 0);
}

WPlanetStructureHeader::~WPlanetStructureHeader()
{ }

void
WPlanetStructureHeader::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setSolidBackground();

    string_t header = planet.getName(GObject::PlainName);
    if (UIBaseWidget* p = group.getFocusedChild()) {
        switch (p->getId()) {
         case 1001:
            header += _(" - Planetary Economy");
            break;
         case 1002:
            header += _(" - Mining Industry");
            break;
         case 1003:
            header += _(" - Defense");
            break;
        }
    }

    ctx.setColor(UIColor::tc_Heading);
    outTextF(ctx, extent.x, extent.y, extent.w, header);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Yellow);
    outTextF(ctx, extent.x, extent.y + 25, extent.w,
             format(_("(Id #%d, %s - %d" "\xC2\xB0" "\x46)"),
                    planet.getId(),
                    getTemperatureName(planet.getTemperature()),
                    planet.getTemperature()));

    can.update(getExtent());
}

void
WPlanetStructureHeader::getLayoutInfo(LayoutInfo& li)
{
    // FIXME: the magic number 25 is copied from WControlScreenHeaderTile
    li.min_size = li.pref_size = GfxPoint(30 * font_ems[FONT_TITLE],
                                          25 + font_heights[FONT_NORMAL]);
    li.flags = li_FlexibleH;
}

void
WPlanetStructureHeader::onResize(GfxRect /*oldpos*/, bool /*by_layout*/)
{
    // Buttons are 25x25
    const GfxRect& x = getExtent();
    btn_left.setExtent(GfxRect(x.x + x.w - 55, x.y, 25, 25), true);
    btn_right.setExtent(GfxRect(x.x + x.w - 25, x.y, 25, 25), true);
}

/******************************* Functions *******************************/

/** Generate build screen 1: General.
    \param action Action to observe
    \param h Place new objects here
    \return Complete build screen page */
static UIBaseWidget&
buildBuildScreen1(GPlanetBuildStructuresAction& action, ObjectHolder& h)
{
    // Build screen 1:
    //   VBox
    //     HBox
    //       VBox
    //         WPlanetStructureCostWidget
    //         Spacer
    //         WPlanetStructureEffectWidget
    //       VBox
    //         WPlanetStructureWidget x 3
    //     HBox: "S-Sell Supplies", "T-Taxes"
    //     HBox: "OK", "Cancel", "H", "Auto", "Goals"

    UIGroup& group1   = h.add(new UIGroup(UIVBoxLayout::instance5, 1001));
    UIGroup& group11  = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& group111 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group112 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group12  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group13  = h.add(new UIGroup(UIHBoxLayout::instance5));

    group111.add(h.add(new WPlanetStructureCostWidget(action, false)));
    group111.add(h.add(new UISpacer()));
    group111.add(h.add(new WPlanetStructureEffectWidget(action)));

    UIBaseWidget& w1 = h.add(new WPlanetStructureGroup(100, action, str_Mines));
    UIBaseWidget& w2 = h.add(new WPlanetStructureGroup(101, action, str_Factories));
    UIBaseWidget& w3 = h.add(new WPlanetStructureGroup(102, action, str_Defense));

    group112.add(w1);
    group112.add(w2);
    group112.add(w3);
    group112.add(h.add(new UISpacer())); // needed to assimilate excess space

    group11.add(group111);
    group11.add(h.add(new UISpacer()));
    group11.add(group112);

    group12.add(h.add(new UIButton(UIKeyString(_("S - Sell Supplies")))));
    group12.add(h.add(new UIButton(UIKeyString(_("T - Taxes")))));
    group12.add(h.add(new UISpacer()));

    group13.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group13.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group13.add(h.add(new UIButton("H", 'h')));
    group13.add(h.add(new UIButton(UIKeyString(_("A - Auto")))));
    group13.add(h.add(new UIButton(UIKeyString(_("G - Goals")))));
    group13.add(h.add(new UISpacer()));

    group1.add(group11);
    group1.add(group12);
    group1.add(group13);
    group1.add(h.add(new UIFocusIterator(100, 102, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Wrap)));

    w1.setFocusedWidget();

    return group1;
}

/** Generate build screen 2: Mining.
    \param action Action to observe
    \param h Place new objects here
    \return Complete build screen page */
static UIBaseWidget&
buildBuildScreen2(GPlanetBuildStructuresAction& action, ObjectHolder& h)
{
    // Build screen 2:
    //   HBox
    //     VBox
    //       WPlanetStructureCostWidget
    //       WPlanetStructureGroup
    //       Spacer
    //       WPlanetStructureEffectWidget
    //       HBox "Sell", "Tax"
    //       HBox "OK", "Cancel", "H", "A", "G"
    //     VBox
    //       Mining forecast x 4
    UIGroup& group1   = h.add(new UIGroup(UIHBoxLayout::instance0, 1002));
    UIGroup& group11  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group111 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group112 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group12  = h.add(new UIGroup(UIVBoxLayout::instance5));

    UIBaseWidget& w1 = h.add(new WPlanetStructureGroup(100, action, str_Mines));

    group11.add(h.add(new WPlanetStructureCostWidget(action, false)));
    group11.add(w1);
    group11.add(h.add(new UISpacer()));
    group11.add(h.add(new WPlanetStructureEffectWidget(action)));
    group11.add(group111);
    group11.add(group112);

    group111.add(h.add(new UIButton(UIKeyString(_("S - Sell Supplies")))));
    group111.add(h.add(new UIButton(UIKeyString(_("T - Taxes")))));
    group111.add(h.add(new UISpacer()));

    group112.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group112.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group112.add(h.add(new UIButton("H", 'h')));
    group112.add(h.add(new UIButton("A", 'a')));
    group112.add(h.add(new UIButton("G", 'g')));
    group112.add(h.add(new UISpacer()));

    group12.add(h.add(new WPlanetStructureBuildMiningTile(action, el_Neutronium)));
    group12.add(h.add(new WPlanetStructureBuildMiningTile(action, el_Tritanium)));
    group12.add(h.add(new WPlanetStructureBuildMiningTile(action, el_Duranium)));
    group12.add(h.add(new WPlanetStructureBuildMiningTile(action, el_Molybdenum)));
    group12.add(h.add(new UISpacer()));

    group1.add(group11);
    group1.add(h.add(new UISpacer()));
    group1.add(group12);

    w1.setFocusedWidget();

    return group1;
}

/** Generate build screen 3: Defense.
    \param action Action to observe
    \param h Place new objects here
    \return Complete build screen page */
static UIBaseWidget&
buildBuildScreen3(GPlanetBuildStructuresAction& action, ObjectHolder& h)
{
    // Build screen 3:
    //   VBox
    //     HBox
    //       VBox
    //         WPlanetStructureCostWidget
    //         Spacer
    //         WPlanetStructureEffectWidget
    //         HBox "Sell" "Tax"
    //       VBox
    //         WPlanetStructureGroup x 2
    //         WPlanetDefenseEffectWidget
    //     HBox "OK" "ESC" "H" "Auto" "Goals" "Ground Combat"

    // FIXME: this forces the WPlanetStructureGroup to the same size
    // as the WPlanetDefenseEffectWidget.

    UIGroup& group1    = h.add(new UIGroup(UIVBoxLayout::instance5, 1003));
    UIGroup& group11   = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& group111  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group1111 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group112  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group12   = h.add(new UIGroup(UIHBoxLayout::instance5));

    group1111.add(h.add(new UIButton(UIKeyString(_("S - Sell Supplies")))));
    group1111.add(h.add(new UIButton(UIKeyString(_("T - Taxes")))));
    group1111.add(h.add(new UISpacer()));

    group111.add(h.add(new WPlanetStructureCostWidget(action, true)));
    group111.add(h.add(new UISpacer()));
    group111.add(h.add(new WPlanetStructureEffectWidget(action)));
    group111.add(group1111);

    UIBaseWidget& w1 = h.add(new WPlanetStructureGroup(100, action, str_Defense));
    UIBaseWidget& w2 = h.add(new WPlanetStructureGroup(101, action, str_BaseDefense));

    group112.add(w1);
    group112.add(w2);
    group112.add(h.add(new UISpacer()));
    group112.add(h.add(new WPlanetDefenseEffectWidget(action)));

    group11.add(group111);
    group11.add(h.add(new UISpacer()));
    group11.add(group112);

    group12.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group12.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group12.add(h.add(new UIButton("H", 'h')));
    group12.add(h.add(new UIButton(UIKeyString(_("A - Auto")))));
    group12.add(h.add(new UIButton(UIKeyString(_("G - Goals")))));
    group12.add(h.add(new UIButton(UIKeyString(_("C - Gnd Combat")))));
    group12.add(h.add(new UISpacer()));

    group1.add(group11);
    group1.add(group12);

    // Is there a base?
    if (!action.getPlanet().hasBase()) {
        // Base defense is disabled if no base
        w2.setState(UIBaseWidget::st_Disabled, true);
    } else {
        // If there is a base, we need a focus iterator
        group112.add(h.add(new UIFocusIterator(100, 101, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Wrap)));
    }
    w1.setFocusedWidget();

    return group1;
}

/** Planet Structure Build Screen: Main Entry Point.
    \param planet Planet to build structures on
    \param start_screen Starting screen: 0=general, 1=mining, 2=defense */
void
doPlanetBuildScreen(GPlanet& planet, int start_screen)
{
    GMultiTransaction mt;
    GPlanetTransfer& transfer =
        mt.addNewSubtransaction(new GPlanetTransfer(planet));
    GPlanetBuildStructuresAction& action =
        mt.addNewSubtransaction(new GPlanetBuildStructuresAction(planet, transfer));

    // FIXME: displayed turn
    action.setUndoInformation(getDisplayedTurn().getPreviousUniverse().getPlanet(planet.getId()));

    ObjectHolder h;             // ATTN: widgets must die before the action

    UIPanel win;
    win.setState(UIWidget::st_Modal, true);
    win.setLayout(UIVBoxLayout::instance5);

    // Window [VBox]
    //   WPlanetStructureHeader
    //   UICardGroup
    //     Screen 1, 2, 3; each of them contains all controls
    UICardGroup& cards = h.add(new UICardGroup());
    cards.add(buildBuildScreen1(action, h));
    cards.add(buildBuildScreen2(action, h));
    cards.add(buildBuildScreen3(action, h));

    win.add(h.add(new WPlanetStructureHeader(planet, cards)));
    win.add(cards);
    win.add(h.add(new UIPrefixArg()));
    win.add(h.add(new WHelpWidget("pcc2:buildings")));
    win.add(h.add(new UIFocusIterator(1001, 1003, UIFocusIterator::fi_Horiz | UIFocusIterator::fi_Tab | UIFocusIterator::fi_Wrap)));
    win.add(h.add(new UIQuit(cm_Escape)));
    cards.setFocusedWidget(*cards.getWidgetById(1001 + start_screen));
    cards.setFocusedWidget();

    win.setExtent(ui_root->getExtent());
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);
    if (rv == cm_Enter && !mt.getError()) {
        mt.doCommit();
    } else {
        mt.doCancel();
    }
}
