/**
  *  \file client/tiles/visheader.cc
  *  \brief Visual Scanner: Header Tile
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/visheader.h"
#include "client/chart.h"
#include "cpluslib/format.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/team.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "client/marks.h"

/************************* WVisualScanHeaderTile *************************/

WVisualScanHeaderTile::WVisualScanHeaderTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(font_sizes[FONT_NORMAL].scaledBy(20, 3).movedBy(0, 3), selection)
{ }

void
WVisualScanHeaderTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);

    GObject* obj = getCurrentObject();
    GShip* s;
    if ((s = dynamic_cast<GShip*>(obj)) != 0) {
        /* it's a ship */
        // Line 1: Ship #1: FOOBAR
        //     or: Ship #1 (guessed position!)
        string_t text;
        if (s->getShipKind() == GShip::GuessedShip)
            text = format(_("Ship #%d (guessed position!)"), s->getId());
        else
            text = s->getName(GObject::LongName);
        outTextF(ctx, 0, 0, extent.w, text);

        if (!s->getAssociatedMessages().empty()) {
            const int SYM_WIDTH = 10;
            int title_width = fonts[FONT_NORMAL_BOLD]->getTextWidth(text) + 5;
            if (title_width < extent.w-SYM_WIDTH) {
                ctx.setColor(UIColor::tc_Blue);
                drawFlag(ctx,
                         GfxPoint(title_width, font_heights[FONT_NORMAL_BOLD] * 7/10),
                         font_heights[FONT_NORMAL_BOLD]*5,
                         font_heights[FONT_NORMAL_BOLD]*3);
            }

        }

        // Line 2: (our <ship>)
        //         (our <ship> under X control)
        //         (X <ship>)
        //         (X <ship> under Y control)
        //         (X <ship> under our control)
        // where <ship> is either 'freighter', 'starship', or 'ship'
        ctx.useFont(*fonts[FONT_NORMAL]);
        enum { Freighter, Starship, Ship } kind;
        if (s->getHullId().isKnown()) {
            if (s->getShipKind() == GShip::CurrentShip) {
                if (s->getNumBeams() > 0 || s->getNumBays() > 0 || s->getNumTorpLaunchers() > 0)
                    kind = Starship;
                else
                    kind = Freighter;
            } else {
                const GHull& h = getHull(s->getHullId());
                if (h.getMaxBeams() > 0 || h.getNumBays() > 0 || h.getMaxLaunchers() > 0)
                    kind = Starship;
                else
                    kind = Freighter;
            }
        } else {
            kind = Ship;
        }

        if (s->getRealOwner() == getPlayerId()) {
            static const char*const table[] = {
                N_("our freighter"),
                N_("our starship"),
                N_("our ship")
            };
            text = _(table[kind]);
        } else {
            static const char*const table[] = {
                N_("%s freighter"),
                N_("%s starship"),
                N_("%s ship")
            };
            text = format(_(table[kind]), player_racenames.getAdjName(s->getRealOwner()));
        }

        if (s->getRealOwner() != s->getOwner()) {
            if (s->getOwner() == getPlayerId()) {
                text += _(" under our control");
            } else {
                text += format(_(" under %s control"), player_racenames.getAdjName(s->getOwner()));
            }
        }

        ctx.setColor(getPlayerColor(s->getOwner()));
        outTextF(ctx, 0, font_heights[FONT_NORMAL], extent.w, "(" + text + ")");

        // Line 3: Unknown type
        //         HULL CLASS
        //         Experienced HULL CLASS
        int16_t level = s->unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
        text.clear();
        if (s->getHullId().isKnown()) {
            if (level >= 0) {
                text = getExperienceLevelName(level);
                text += ' ';
            }
            text += getHull(s->getHullId()).getName();
        } else {
            if (level >= 0) {
                text = getExperienceLevelName(level);
            } else {
                text = _("Unknown type");
            }
        }
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, 0, 2*font_heights[FONT_NORMAL]+3, extent.w, text);
    } else {
        /* something else */
        if (obj != 0)
            outTextF(ctx, 0, 0, extent.w, obj->getName(GObject::PlainName));
        else
            outTextF(ctx, 0, 0, extent.w, string_t());
        outTextF(ctx, 0, font_heights[FONT_NORMAL], extent.w, string_t());
        outTextF(ctx, 0, 2*font_heights[FONT_NORMAL]+3, extent.w, string_t());
    }
}

