/**
  *  \file client/widgets/colorselect.cc
  *  \brief WColorSelector widget
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/colorselect.h"
#include "cpluslib/assert.h"
#include "game/drawing.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/widgets/quit.h"

/***************************** WColorSelector ****************************/

enum {
    CS_LINES   = 3,
    CS_COLUMNS = 10
};
CTASSERT(CS_LINES * CS_COLUMNS == GDrawing::NUM_USER_COLORS);

/** Constructor. */
WColorSelector::WColorSelector()
    : user_color(1)
{ }

/** Destructor. */
WColorSelector::~WColorSelector()
{ }

/** Move cursor by specified delta. */
void
WColorSelector::moveBy(int delta)
{
    int n = user_color + delta;
    if (n <= 0)
        n += GDrawing::NUM_USER_COLORS;
    if (n > GDrawing::NUM_USER_COLORS)
        n -= GDrawing::NUM_USER_COLORS;
    setColor(n);
}

/** Get size of one color box. */
int
WColorSelector::getFrameSize() const
{
    return font_heights[FONT_TITLE];
}

/** Select a color.
    \param user_color User color index. */
void
WColorSelector::setColor(int user_color)
{
    if (user_color >= 0 && user_color <= GDrawing::NUM_USER_COLORS) {
        if (user_color != this->user_color) {
            this->user_color = user_color;
            drawWidget();
        }
    }
}

/** Get selected color. */
int
WColorSelector::getColor() const
{
    return user_color;
}

void
WColorSelector::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(CS_COLUMNS*getFrameSize(), CS_LINES*getFrameSize());
}

void
WColorSelector::drawContent(GfxCanvas& can)
{
    const int fs = getFrameSize();
    for (int line = 0; line < CS_LINES; ++line) {
        for (int col = 0; col < CS_COLUMNS; ++col) {
            const int this_color = 1 + col + line*CS_COLUMNS;
            const int x = getExtent().x + col*fs;
            const int y = getExtent().y + line*fs;
            // Draw frame
            if (this_color == user_color) {
                drawFrameDown(can, GfxRect(x, y, fs, fs));
            } else {
                drawFrameUp(can, GfxRect(x, y, fs, fs));
            }

            // Draw color box
            drawSolidBar(can, GfxRect(x+1, y+1, fs-2, fs-2),
                         standard_colors[GDrawing::getColorFromUserColor(this_color)]);
        }
    }
    can.update(getExtent());
}

bool
WColorSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse && e.mouse.b != 0 && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        // Mouse clicked inside widget
        setSelectedWidget();
        int col  = (e.mouse.x - getExtent().x) / getFrameSize();
        int line = (e.mouse.y - getExtent().y) / getFrameSize();
        if (col < CS_COLUMNS && line < CS_LINES)
            setColor(1 + col + line*CS_COLUMNS);
        return true;
    }
    if (e.type == ev_Key) {
        // Handle some keys
        switch (e.key.code) {
         case SDLK_LEFT:
            setSelectedWidget();
            moveBy(-1);
            return true;
         case SDLK_RIGHT:
            setSelectedWidget();
            moveBy(+1);
            return true;
         case SDLK_UP:
            setSelectedWidget();
            moveBy(-10);
            return true;
         case SDLK_DOWN:
            setSelectedWidget();
            moveBy(+10);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

/** Perform standard dialog.
    \param title Window title
    \param offer_adjacent true to offer "Adjacent" button
    \return rCancel, rOK, rAdjacent */
int
WColorSelector::doStandardDialog(string_t title, bool offer_adjacent)
{
    ObjectHolder h;
    UIWindow w(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    w.setLayout(UIVBoxLayout::instance5);
    w.add(h.add(new UIStaticText(_("Choose color:"), UIStaticText::StaticTitle)));
    w.add(*this);
    g.add(h.add(new UIButton(rOK, _("OK"), SDLK_RETURN)).addStop());
    if (offer_adjacent)
        g.add(h.add(new UIButton(rAdjacent, _("Adjacent"), 'a')).addStop());
    g.add(h.add(new UIButton(rCancel, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    w.add(g);
    w.add(h.add(new UIQuit(rCancel)));
    w.pack();
    ui_root->centerWidget(w);
    return ui_root->runModal(w);
}
