/**
  *  \file game/msg.h
  *  \brief Subspace Message Handling
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_MSG_H
#define PCC_V2_GAME_MSG_H

#include <vector>
#include <list>

#include "cpluslib/sigslot.h"
#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "game/playerset.h"
#include "cpluslib/ptrvector.h"

//! Subspace message box.
class GMessagebox {
    int current;
 public:
    GMessagebox();
    virtual ~GMessagebox();

    /// \name Message Interface
    //@{
    /** Get number of messages in this message box. */
    virtual int getCount() = 0;
    /** Get text of a message.
        \param index message number, [0, getCount()). */
    virtual string_t getText(int index) = 0;
    /** Get heading of a message. The heading is used for
        subject-sorting and the kill filter.
        \param index message number, [0, getCount()). */
    virtual string_t getHeading(int index) = 0;
    //@}

    /// \name Message Pointer
    //@{
    /** Get currently-selected message. */
    int getCurrent() const
        { return current; }
    void setCurrent(int aindex);
    /** Raised when currently-selected message changes. */
    Signal0 sig_sel_changed;
    //@}
};

//! Standard Outbox.
class GOutbox : public GMessagebox {
 public:
    enum Format {
        fDos,                   ///< DOS-style mailbox (messX.dat).
        fWinplan                ///< Winplan-style mailbox (mess35X.dat).
    };
 private:
    struct Message;
    ptr_vector<Message> messages;
    int sender;
    Format format;
 public:
    GOutbox(int sender);
    ~GOutbox();

    int      getCount();
    string_t getText(int index);
    string_t getHeading(int index);
    string_t getSendPrefix(int index, int to);
    string_t getRawText(int index);
    GPlayerSet getReceiverMask(int index);

    void setReceiverMask(int index, GPlayerSet receivers);
    void setMessage(int index, string_t text, GPlayerSet receivers);

    void     deleteMessagesAfter(int index);
    void     deleteMessage(int index);

    /** Get sender of this mailbox. */
    int getSender() const
        { return sender; }
    /** Set mailbox storage format. */
    void setFormat(Format format)
        { this->format = format; }
    /** Get mailbox storage format. */
    Format getFormat() const
        { return format; }

    void     addMessage(string_t text, GPlayerSet receivers);
    void     addMessageFromFile(string_t text, GPlayerSet receivers);

    /** All permitted receivers (Host + 11 players) */
    static const uint16 ALL_RECEIVERS_MASK = 0x1FFE;
    /** Receivers of a standard universal message (11 players) */
    static const uint16 UNIVERSAL_MASK     = 0x0FFE;
    static const char   UNIVERSAL_TEXT[];

    string_t getHeadersForDisplay(GPlayerSet receivers);
};

//! Message filter.
class GMessageFilter {
    std::list<string_t> filter;

    typedef std::list<string_t>::iterator entryptr_t;

    entryptr_t getEntry(const string_t& header);
 public:
    typedef std::list<string_t>::const_iterator iterator;

    GMessageFilter();
    ~GMessageFilter();

    bool isFiltered(const string_t& header) const;
    bool toggleFilter(const string_t& header);
    void addToFilter(const string_t& header);
    void removeFromFilter(const string_t& header);
    void clear();
    bool empty() const;
    iterator begin() const;
    iterator end() const;
};

// Global Message Configuration
void initMessageConfig();
void doneMessageConfig();
void loadMessageConfig(int player);
void saveMessageConfig(int player);
GMessageFilter& getMessageFilter();

string_t getMessageFromArray(const char* array, std::size_t size, bool rewrap);
void     encodeMessage(string_t& text);

extern const struct ModuleInitializer mod_msg[];

#endif
