/**
  *  \file game/sel.cc
  *  \brief Selection Layers
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "game/sel.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/univ.h"
#include "util/except.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "game/actions/preconditions.h"

/******************************* GSelection ******************************/

/** Constructor. */
GSelection::GSelection()
{
    clear();
}

/** Clear this selection layer. */
void
GSelection::clear()
{
    zeroFill(data);
}

/** Initialize this selection layer from a universe.
    \param univ Universe to copy from */
void
GSelection::copyFrom(const GUniverse& univ)
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid)
        set(FIRST_SHIP_INDEX + sid-1, univ.getShip(sid).isMarked());
    for (int pid = 1; pid <= NUM_PLANETS; ++pid)
        set(FIRST_PLANET_INDEX + pid-1, univ.getPlanet(pid).isMarked());
}

/** Copy this selection layer to a universe. The universe's selection status will reflect this layer.
    \param univ Universe to copy to */
void
GSelection::copyTo(GUniverse& univ)
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid)
        univ.getShip(sid).setIsMarked(get(FIRST_SHIP_INDEX + sid-1));
    for (int pid = 1; pid <= NUM_PLANETS; ++pid)
        univ.getPlanet(pid).setIsMarked(get(FIRST_PLANET_INDEX + pid-1));
}

/** Limit to existing objects. Unmarks all objects that do not exist.
    - Ships: history ships are treated as existing, even if they are not visible now.
      This means it's possible for the selection to be nonempty, but no marker on the
      starchart
    - Planets: all planets are treated as existing unless they're out of the map */
void
GSelection::limitToExistingObjects(const GUniverse& univ)
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid)
        if (!univ.ty_history_ships.isValidIndex(sid))
            set(FIRST_SHIP_INDEX + sid-1, false);
    for (int pid = 1; pid <= NUM_PLANETS; ++pid)
        if (!univ.ty_any_planets.isValidIndex(pid))
            set(FIRST_PLANET_INDEX + pid-1, false);
}

/** Get number of ships in this selection. */
int
GSelection::getNumShips() const
{
    int n = 0;
    for (int sid = 1; sid <= NUM_SHIPS; ++sid)
        if (get(FIRST_SHIP_INDEX + sid-1))
            ++n;
    return n;
}

/** Get number of planets in this selection. */
int
GSelection::getNumPlanets() const
{
    int n = 0;
    for (int pid = 1; pid <= NUM_PLANETS; ++pid)
        if (get(FIRST_PLANET_INDEX + pid-1))
            ++n;
    return n;
}

// Bitwise access

/** Get selection status of an object.
    \param index 0-based index of object */
bool
GSelection::get(uint32_t index) const
{
    uint32_t word_index = index / 32;
    uint32_t bit_index = index % 32;
    return (data[word_index] & (1 << bit_index)) != 0;
}

/** Set selection status of an object.
    \param index 0-based index of object
    \param value select or unselect this object */
void
GSelection::set(uint32_t index, bool value)
{
    uint32_t word_index = index / 32;
    uint32_t bit_index = index % 32;
    if (value)
        data[word_index] |= 1 << bit_index;
    else
        data[word_index] &= ~(1 << bit_index);
}

// Word-wise access

/** Get word (encoded multiple objects).
    \param word_index 0-based word index */
uint32_t
GSelection::operator[](uint32_t word_index) const
{
    return data[word_index];
}

/** Get word (encoded multiple objects) with write access.
    \param word_index 0-based word index */
uint32_t&
GSelection::operator[](uint32_t word_index)
{
    return data[word_index];
}

/** Get ship mask for a word index. A nonzero bit in the return value means that
    the same bit in operator[](word_index) corresponds to a ship.
    \param word_index 0-based word index */
uint32_t
GSelection::getShipMask(uint32_t word_index)
{
    CTASSERT(FIRST_SHIP_INDEX < FIRST_PLANET_INDEX);

    /* A really simple implementation: */
    uint32_t bit_index = word_index * 32;
    uint32_t result = 0;
    for (int i = 0; i < 32; ++i, ++bit_index)
        if (bit_index < FIRST_PLANET_INDEX)
            result |= (1 << i);
    return result;
}

/** Get planet mask for a word index. A nonzero bit in the return value means that
    the same bit in operator[](word_index) corresponds to a planet.
    \param word_index 0-based word index */
uint32_t
GSelection::getPlanetMask(uint32_t word_index)
{
    return ~getShipMask(word_index);
}

/**************************** GMultiSelection ****************************/

static int current_selection_layer;
static GSelection selection_layers[GMultiSelection::NUM_SELECTION_LAYERS];
Signal0 GMultiSelection::sig_selection_changed;

static void
selectionError()
{
    throw GError("Invalid selection operation", _("Invalid selection operation"));
}

static void
clearSelections()
{
    for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i)
        GMultiSelection::getSelectionLayer(i).clear();
    GMultiSelection::sig_selection_changed.raise();
}

/** Get current selection layer.
    \return [0,NUM_SELECTION_LAYERS) */
int
GMultiSelection::getCurrentSelectionLayer()
{
    return current_selection_layer;
}

/** Set current selection layer. Modifies the currently displayed turn
    to reflect the change.
    \param layer New layer, [0,NUM_SELECTION_LAYERS) */
void
GMultiSelection::setCurrentSelectionLayer(int layer)
{
    ASSERT(isValidSelectionLayer(layer));
    if (layer != current_selection_layer) {
        if (haveDisplayedTurn()) {
            // We have a turn, change with copying.
            GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
            getSelectionLayer(current_selection_layer).copyFrom(univ);
            current_selection_layer = layer;
            getSelectionLayer(current_selection_layer).copyTo(univ);
            getSelectionLayer(current_selection_layer).limitToExistingObjects(univ);
        } else {
            // We do not have a turn. Just register that we now have a new layer.
            current_selection_layer = layer;
        }
        sig_selection_changed.raise();
    }
}

/** Evaluate selection expression.
    \param target_layer Layer to modify, [0,NUM_SELECTION_LAYERS)
    \param compiled_expression Compiled expression. The string contains opXXX. See int/selexpr.h for the compiler. */
void
GMultiSelection::executeSelectionExpression(const int target_layer, const string_t& compiled_expression)
{
    ASSERT(isValidSelectionLayer(target_layer));
    mustHaveTurn();
    getSelectionLayer(current_selection_layer).copyFrom(getDisplayedTurn().getCurrentUniverse());

    for (uint32_t i = 0; i < GSelection::NUM_WORDS; ++i) {
        std::vector<uint32_t> stack;
        for (string_t::size_type x = 0; x < compiled_expression.size(); ++x) {
            switch (compiled_expression[x]) {
             case opAnd:
                if (stack.size() < 2)
                    selectionError();
                stack[stack.size()-2] &= stack[stack.size()-1];
                stack.pop_back();
                break;
             case opOr:
                if (stack.size() < 2)
                    selectionError();
                stack[stack.size()-2] |= stack[stack.size()-1];
                stack.pop_back();
                break;
             case opXor:
                if (stack.size() < 2)
                    selectionError();
                stack[stack.size()-2] ^= stack[stack.size()-1];
                stack.pop_back();
                break;
             case opNot:
                if (stack.size() < 1)
                    selectionError();
                stack[stack.size()-1] = ~stack[stack.size()-1];
                break;
             case opCurrent:
                stack.push_back(getSelectionLayer(current_selection_layer)[i]);
                break;
             case opShip:
                stack.push_back(GSelection::getShipMask(i));
                break;
             case opPlanet:
                stack.push_back(GSelection::getPlanetMask(i));
                break;
             case opZero:
                stack.push_back(0);
                break;
             case opOne:
                stack.push_back(~0U);
                break;
             default:
                if (compiled_expression[x] >= opFirstLayer && compiled_expression[x] < opFirstLayer + NUM_SELECTION_LAYERS)
                    stack.push_back(getSelectionLayer(compiled_expression[x] - opFirstLayer)[i]);
                else
                    selectionError();
                break;
            }
        }
        if (stack.size() != 1)
            selectionError();
        selection_layers[target_layer][i] = stack[0];
    }

    selection_layers[target_layer].limitToExistingObjects(getDisplayedTurn().getCurrentUniverse());
    if (target_layer == current_selection_layer)
        getSelectionLayer(current_selection_layer).copyTo(getDisplayedTurn().getCurrentUniverse());

    sig_selection_changed.raise();
}

/** Get selection layer by number.
    \param layer Layer [0,NUM_SELECTION_LAYERS) */
GSelection&
GMultiSelection::getSelectionLayer(int layer)
{
    ASSERT(isValidSelectionLayer(layer));
    return selection_layers[layer];
}

/** Check for valid selection layer.
    \param layer Value to check */
bool
GMultiSelection::isValidSelectionLayer(int layer)
{
    return (layer >= 0 && layer < NUM_SELECTION_LAYERS);
}

/***************************** Initialisation ****************************/

const ModuleInitializer mod_sel[] = {
    { OnEnterNoGame | OnEnterGame | OnLeftGame | OnEnterDirectory, clearSelections },
    { 0, 0 },
};
