/**
  *  \file int/if/guiif.cc
  *  \brief Interpreter Game Interface - GUI Components
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <climits>
#include <memory>
#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dialogs/allydlg.h"
#include "client/dialogs/baseammo.h"
#include "client/dialogs/baseshipbuild.h"
#include "client/dialogs/buildqueue.h"
#include "client/dialogs/cargohist.h"
#include "client/dialogs/explainpredict.h"
#include "client/dialogs/helpdlg.h"
#include "client/dialogs/histselect.h"
#include "client/dialogs/inboxdlg.h"
#include "client/dialogs/keymapdlg.h"
#include "client/dialogs/multitransfer.h"
#include "client/dialogs/navchart.h"
#include "client/dialogs/prefdlg.h"
#include "client/dialogs/resetdlg.h"
#include "client/dialogs/searchdlg.h"
#include "client/dialogs/selmgr.h"
#include "client/dialogs/shipcostcalc.h"
#include "client/dialogs/simplemboxdlg.h"
#include "client/dlg-fcode.h"
#include "client/dlg-mission.h"
#include "client/dlg-structure.h"
#include "client/dlg-supplies.h"
#include "client/dlg-tax.h"
#include "client/dlg-tech.h"
#include "client/dlg-visual.h"
#include "client/dlg-warp.h"
#include "client/history.h"
#include "client/remote.h"
#include "client/screens.h"
#include "client/screens/basescreen.h"
#include "client/screens/fleetscreen.h"
#include "client/screens/planetscreen.h"
#include "client/screens/shipscreen.h"
#include "client/widgets/fleetlist.h"
#include "client/widgets/objlist.h"
#include "client/widgets/specview.h"
#include "cpluslib/format.h"
#include "cpluslib/utf8.h"
#include "game/actions/preconditions.h"
#include "game/actions/preconditions.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/fleet.h"
#include "game/gamedirs.h"
#include "game/global.h"
#include "game/host.h"
#include "game/mission.h"
#include "game/planet.h"
#include "game/proxymbox.h"
#include "game/racename.h"
#include "game/reg.h"
#include "game/reset.h"
#include "game/ship.h"
#include "game/shipquery.h"
#include "game/simplembox.h"
#include "game/team.h"
#include "game/utilmsg.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/exec.h"
#include "int/if/guiif.h"
#include "int/if/ifutil.h"
#include "int/if/listif.h"
#include "int/if/plugif.h"
#include "int/if/richif.h"
#include "int/if/userif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "resmgr/resmgr.h"
#include "ui/commands.h"
#include "ui/dialogs/filedlg.h"
#include "ui/event.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/inputline.h"
#include "util/helpindex.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/strlist.h"

namespace {
    class IntFileDialogDirectoryUpdater {
     public:
        IntFileDialogDirectoryUpdater(UIFileChooserDialog& dlg, IntExecutionContext& exc)
            : dlg(dlg),
              exc(exc)
            { }
        void update();

     private:
        UIFileChooserDialog& dlg;
        IntExecutionContext& exc;
    };

    string_t getPluginContext(IntExecutionContext& exc)
    {
        string_t name = "Script";
        const ptr_vector<IntContext>& ctxs = exc.getContexts();
        for (size_t i = ctxs.size(); i > 0; --i) {
            if (IntPluginContext* plugContext = dynamic_cast<IntPluginContext*>(ctxs[i-1])) {
                name = plugContext->getId();
                break;
            }
        }
        return name;
    }
}

void
IntFileDialogDirectoryUpdater::update()
{
    Ptr<Directory> dir = dlg.getCurrentDirectory();
    if (dir.ptr()) {
        IntStringValue sv(dir->getName());
        exc.setVariable("UI.DIRECTORY", &sv);
    }
}

void
flushUI()
{
    // Tell widgets to order their redraws
    if (haveDisplayedTurn())
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    // Perform redraws
    if (ui_root)
        ui_root->performDeferredRedraws();
}

/* @q MessageBox text:Str, Optional heading:Str (Global Command)
   Display a message.
   In the graphical interface, displays an "OK" message box.
   In console mode, just prints out a message and continues.
   @see UI.Message
   @since PCC 1.0.6, PCC2 1.99.9 */
static void
IFMessageBox(IntExecutionContext& exc, IntArgBlock& args)
{
    /* MessageBox <text>[, <heading>] */

    // Validate arguments
    args.checkArgumentCount(1, 2);

    string_t content;
    string_t heading = _("Message");
    if (!checkStringArg(content, args.getNext()))
        return;
    checkStringArg(heading, args.getNext());

    // Do it
    flushUI();
    if (exc.checkForBreak())
        return;
    messageBox(content, heading);
}

/* @q UI.Update Optional flag:Bool (Global Command)
   Update graphical user interface.
   This causes all the screen to be redrawn.
   With the flag specified as %True, redraws even if there are no changes.

   In console mode, this function does nothing.
   @since PCC 1.0.13, PCC2 1.99.9 */
static void
IFUIUpdate(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* UI.Update [<forceFullRedrawFlag>] */

    args.checkArgumentCount(0, 1);

    int mode = 0;
    if (args.getNumArgs() > 0)
        mode = getBoolValue(args.getNext());

    if (mode >= 0) {
        // Regular redraw
        flushUI();

        // Force full redraw
        if (mode > 0 && ui_root)
            ui_root->drawWidget(true);
    }
}

/* @q UI.FileWindow title:Str, wildcard:Str, Optional helpId:Str (Global Command)
   File selection.

   Opens  the "select a file" dialog and lets the user choose a file.
   The %title argument specifies what to show in the window title,
   the %wildcard is a wildcard which specifies the default filter.
   For example, to choose a log file, do
   | UI.FileWindow "Choose Log File", "*.log"

   The optional third argument specifies a help page to use,
   it defaults to the help page for the file window.
   See {UI.Help} for more information.

   When the user hits "OK", this command returns the chosen file in
   {UI.Result}; when the user cancels, UI.Result is set to EMPTY.

   The file dialog uses the variable {UI.Directory} to initialize and store the current directory.

   In text mode, this command gives a simple, no-frills input line ({UI.Input}).

   @c This command also accepts a fourth argument that deals with
   @c predefined file names; details are still undocumented at this place.

   @diff In PCC 1.x, the help Id is an integer. In PCC2, it is a string.
   @since PCC2 1.99.21, PCC 1.0.15 */
static void
IFUIFileWindow(IntExecutionContext& exc, IntArgBlock& args)
{
    /* Parse args */
    args.checkArgumentCount(2, 3);

    string_t title, wildcard, helpId;
    if (!checkStringArg(title, args.getNext()) || !checkStringArg(wildcard, args.getNext())) {
        return;
    }
    checkStringArg(helpId, args.getNext());

    /* UI cruft */
    flushUI();
    if (exc.checkForBreak())
        return;

    /* Get current directory */
    string_t dirName = IntValue::toStringOf(exc.getVariable("UI.DIRECTORY"), false);
    if (dirName.empty()) {
        dirName = Directory::getWorkingDirectory();
    }

    /* Prepare dialog */
    UIFileChooserDialog dlg(title, Directory::openDir(dirName), wildcard);
    IntFileDialogDirectoryUpdater updater(dlg, exc);
    dlg.getDirectoryChangeSignal().add(&updater, &IntFileDialogDirectoryUpdater::update);

    /* Do it */
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg)) {
        IntStringValue sv(dlg.getResultFile());
        exc.setVariable("UI.RESULT", &sv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.Input prompt:Str, Optional title:Str, max:Int, flags:Any, def:Str (Global Command)
   Text input.

   Displays a standard text input dialog.
   All but the first parameter are optional, and have the following meaning:
   - %prompt: the prompt to show in the dialog.
   - %title: the window title. Defaults to the value of %prompt.
   - %length: maximum number of characters to accept, defaults to 255.
   - %flags: some additional flags that affect the behaviour of this dialog (see below).
   - %def: initial contents of text input field.

   The %flags parameter is a string that can contain the following:
   - "n": only accept numeric input (digits).
   - "h": do not accept high-ASCII characters (umlauts, accents, etc.).
   - "p": password input: display stars instead of actual input.
   - "f": draw a frame around the input line.
   - "g": only accept characters from game character set.
   - "m": the width (next item) is specified in ems (default: pixels).
   - a number: width of input line.
  For example, "h450" gives an input line which is 450 pixels wide and does not accept high-ASCII input,
  "30m" displays an input line which is 30 ems wide.
  If you only want to set the width, you can also pass an integer instead of a string.

  The result is returned in {UI.Result}:
  if the user hits <kbd>Enter</kbd>, {UI.Result} contains the input.
  If the user hits <kbd>ESC</kbd>, {UI.Result} will be EMPTY.

  Example: this is the "rename ship" function <kbd>N</kbd> on the ship screen:
  | UI.Input "Enter new name:", "Rename Starship #" & Id, 20, 320, Name
  | SetName UI.Result
  ({SetName (Ship Command)|SetName} will not do anything when passed an EMPTY parameter).

  In text mode, this command makes a simple input line using the %prompt only.
  @since PCC 1.0.9, PCC2 1.99.9 */
static void
IFUIInput(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Input <prompt>[, <title>, <maxChars>, <flags>, <default>] */
    args.checkArgumentCount(1, 5);

    string_t prompt;
    string_t title;
    string_t defaultText;
    int32_t  maxChars = 255;
    int32_t  flags = 0;
    int32_t  width = ui_root->getExtent().w / 2;

    // Mandatory argument
    if (!checkStringArg(prompt, args.getNext()))
        return;

    // Optional arguments
    title = prompt;
    checkStringArg(title, args.getNext());
    checkIntArg(maxChars, args.getNext(), 0, 32000);
    checkFlagArg(flags, &width, args.getNext(), "NHPFGM");
    checkStringArg(defaultText, args.getNext());

    // Convert flags
    //   N = numeric
    //   H = on high ASCII
    //   P = password masking
    //   F = frame
    //   G = game charset
    //   M = width is in ems
    uint32_t uiFlags = 0;
    if (flags & 1)
        uiFlags |= UIInputLine::ilf_NumbersOnly;
    if (flags & 2)
        uiFlags |= UIInputLine::ilf_NoHi;
    if (flags & 4)
        uiFlags |= UIInputLine::ilf_Hidden;
    if (flags & 8)
        uiFlags |= UIInputLine::ilf_Framed;
    if (flags & 16)
        uiFlags |= UIInputLine::ilf_GameChars;

    if ((flags & 32) == 0 && font_ems[FONT_TITLE] > 0) {
        // width given in in pixels. Convert to ems.
        width /= font_ems[FONT_TITLE];
    }

    // Build a UIInputLine object
    flushUI();
    if (exc.checkForBreak())
        return;
    if (UIInputLine::run(defaultText, maxChars, title, prompt, width, uiFlags)) {
        IntStringValue sv(defaultText);
        exc.setVariable("UI.RESULT", &sv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.InputNumber title:Str, Optional min:Int, max:Int, current:Int, help:Any (Global Command)
   Number input.
   This command prompts for a number, using the standard number input window.

   The parameters are
   - %title: the title/prompt shown in the dialog.
   - %min,%max: the acceptable range of numbers (defaults to 0..10000).
   - %current: the current value (defaults to 0).
   - %help: help page to associate with dialog.
   Only the first parameter is mandatory.

   The result will be returned in {UI.Result}.
   It will be an integer within the requested range, or EMPTY if the user canceled the dialog.

   For example, to change a ship's warp factor, you could use
   | UI.InputNumber "Warp", 0, 9, Speed$
   | SetSpeed UI.Result
   (Note that {SetSpeed (Ship Command)|SetSpeed} is implicitly ignored if its parameter is EMPTY).

   This command currently does not work in text mode.

   @since PCC 1.1.16, PCC2 1.99.9 */
static void
IFUIInputNumber(IntExecutionContext& exc, IntArgBlock& args)
{
    // Check arguments
    args.checkArgumentCount(1, 5);

    string_t title;
    if (!checkStringArg(title, args.getNext()))
        return;

    int32_t min = 0;
    checkIntArg(min, args.getNext());

    int32_t max = 10000;
    checkIntArg(max, args.getNext());

    int32_t current = 0;
    checkIntArg(current, args.getNext());

    string_t help;
    checkStringArg(help, args.getNext());

    // Adjust arguments
    if (max < min)
        std::swap(min, max);

    // Do it.
    flushUI();
    if (exc.checkForBreak())
        return;
    UIDecimalSelector nsel(0, min, max, 10);
    nsel.setValue(current);  /* value will be clipped into valid range */
    if (nsel.doStandardDialog(title, title, help)) {
        IntIntIntValue iv(nsel.getValue());
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.InputFCode flags:Any, Optional default:Str (Global Command)
   Ask for friendly code input.
   This uses the regular friendly code input window with a list of friendly code.

   The %flags parameter is a string that can contain the following options:
   - <tt>"S"</tt>: offer ship friendly codes
   - <tt>"P"</tt>: offer planet friendly codes
   - <tt>"B"</tt>: offer starbase friendly codes
   - <tt>"C"</tt>: offer friendly codes for capital ships
   - <tt>"A"</tt>: offer friendly codes for alchemy ships
   - <tt>"D"</tt>: offer default context-dependant selection.
     This examines the current context and offers matching codes.
     That is, when this command is called from a ship, offers matching ship codes.
     All other options are ignored in this case.
   - a number: offer friendly codes available to the specified player.
     Defaults to the currently loaded player if omitted or zero.

   You should specify either "D", or at least one of "S", "B" and "P".

   The optional %default parameter specifies the current value of the friendly code.
   The code starts as empty if this argument is omitted.

   The result will be stored in %UI.Result, as usual for user interface commands.
   If the dialog is canceled, %UI.Result will be EMPTY.

   In text mode, this command yields a simple input line, like this:
   | UI.Input "Friendly Code", "", 3, "h", default

   @diff The "D" flag is supported in PCC2 only.
   @since PCC2 1.99.21, PCC 1.0.17 */
static void
IFUIInputFCode(IntExecutionContext& exc, IntArgBlock& args)
{
    enum {
        ShipFlag       = 1,
        PlanetFlag     = 2,
        BaseFlag       = 4,
        CapitalFlag    = 8,
        AlchemyFlag    = 16,
        DefaultFlag    = 32,

        // 32 is actually the "Registered" flag but we cannot reasonably deal with it.

        TypeFlags     = ShipFlag | PlanetFlag | BaseFlag,
        PropertyFlags = CapitalFlag | AlchemyFlag
    };
    CTASSERT(int(ShipFlag)       == int(GFCode::fc_Ship));
    CTASSERT(int(PlanetFlag)     == int(GFCode::fc_Planet));
    CTASSERT(int(BaseFlag)       == int(GFCode::fc_Base));
    CTASSERT(int(CapitalFlag)    == int(GFCode::fc_Capital));
    CTASSERT(int(AlchemyFlag)    == int(GFCode::fc_Alchemy));

    // GUI boilerplate
    flushUI();
    if (exc.checkForBreak())
        return;

    // Check arguments
    args.checkArgumentCount(1, 2);

    int flags = 0, player = 0;
    if (!checkFlagArg(flags, &player, args.getNext(), "SPBCAD")) {
        return;
    }
    if (player < 0 || player > NUM_PLAYERS) {
        throw IntError::rangeError();
    }
    if (player == 0) {
        player = ::getPlayerId();
    }

    string_t init;
    checkStringArg(init, args.getNext());

    // Construct a friendly-code list
    std::auto_ptr<GFCodeList> list;
    if ((flags & DefaultFlag) != 0) {
        // Default mode
        GObject* obj = exc.getCurrentObject();
        if (obj == 0) {
            throw IntError::contextError();
        }
        list.reset(new GFCodeList(GFCodeList::getInstance(), *obj));
    } else {
        // Parameterized mode
        list.reset(new GFCodeList());
        const GFCodeList& g = GFCodeList::getInstance();
        for (GFCodeList::iterator gi = g.begin(); gi != g.end(); ++gi) {
            /* An fcode is accepted if
               - flags have ANY of the TypeFlags required by the code
               - flags have ALL of the PropertyFlags required by the code */
            const int fcFlags = gi->getFlags();
            if ((fcFlags & flags & TypeFlags) != 0
                && (fcFlags & ~flags & PropertyFlags) == 0
                && ((fcFlags & GFCode::fc_Registered) == 0 || allowRegistered(player))
                && gi->getRaceMask().contains(player))
            {
                list->addFCode(*gi);
            }
        }
        list->sort();
    }

    // Dialog
    WFCodeWindow dlg(_("Change Friendly Code"), *list);
    dlg.setFCode(init);
    ui_root->centerWidget(dlg);
    ui_root->add(dlg);
    if (dlg.run() == cm_Enter) {
        IntStringValue sv(dlg.getFCode());
        exc.setVariable("UI.RESULT", &sv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.ChooseObject screen:Int (Global Command)
   Choose game object.
   You specify a screen number to choose the object for:
   <table>
    <tr><td width="4">1, 11</td><td width="10">Own starships</td></tr>
    <tr><td width="4">2, 12</td><td width="10">Own planets</td></tr>
    <tr><td width="4">3, 13</td><td width="10">Own starbases</td></tr>
    <tr><td width="4">6</td>    <td width="10">History starships</td></tr>
    <tr><td width="4">10</td>   <td width="10">Fleets</td></tr>
   </table>

  For example, <tt>UI.ChooseObject 1</tt> does the same as the <kbd>F1</kbd> key in most screens.
  When there is just one ship, no dialog is displayed.

  The chosen object Id is returned in {UI.Result}; the result is
  EMPTY if the user canceled.
  This command does not work in text mode.

  @since PCC 1.1.1, PCC2 1.99.9 */
static void
IFUIChooseObject(IntExecutionContext& exc, IntArgBlock& args)
{
    // Check arguments
    int32_t screen;
    args.checkArgumentCount(1);
    if (!checkIntArg(screen, args.getNext(), 0, 100))
        return;

    // Do it
    flushUI();
    if (exc.checkForBreak())
        return;

    int result;
    switch (screen) {
     case 1:
     case 11:
        result = chooseShip();
        break;
     case 2:
     case 12:
        result = choosePlanet();
        break;
     case 3:
     case 13:
        result = chooseBase();
        break;
     case 6:
        /* This needs a GPoint. iuiChartX is always valid even on the player screen,
           unless we're on a control screen of an object that has no position, where
           it returns EMPTY. We want it to be 0 for invalid positions, and on the player
           screen, where there is no user-perceived valid position. */
        /* For now, we take the position from the UI. Taking it from the current object,
           if any, would need to go through the trouble of determining whether the
           current object actually has a position. */
        {
            int currentShipId = 0;
            if (GShip* ps = dynamic_cast<GShip*>(exc.getCurrentObject())) {
                currentShipId = ps->getId();
            }

            IntUserInterfaceBinding& ui = IntUserInterfaceBinding::get();
            std::auto_ptr<IntValue> vx(ui.getProperty(iuiChartX));
            std::auto_ptr<IntValue> vy(ui.getProperty(iuiChartY));
            std::auto_ptr<IntValue> vs(ui.getProperty(iuiScreenNumber));
            int32_t x, y, screen;
            if (checkIntArg(x, vx.get()) && checkIntArg(y, vy.get()) && checkIntArg(screen, vs.get()) && screen > 0) {
                result = chooseHistoryShip(GPoint(x, y), currentShipId);
            } else {
                result = chooseHistoryShip(GPoint(0, 0), currentShipId);
            }
        }
        break;
     case 10:
        result = chooseFleet();
        break;
     default:
        throw IntError::rangeError();
    }

    if (result != 0) {
        IntIntIntValue iv(result);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.EditAlliances (Global Command)
   Alliance editor dialog.
   Brings up a dialog that allows the user to <a href="pcc2:allies">edit alliances</a>.
   This command takes no further parameters.

   @since PCC2 1.99.23 */
static void
IFUIEditAlliances(IntExecutionContext& exc, IntArgBlock& args)
{
    // Preconditions
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();
    if (exc.checkForBreak()) {
        return;
    }

    // Do it
    doAllianceDialog();
}

/* @q UI.Message text:RichText, Optional title:Str, buttons:Str (Global Command)
   Display a message.
   This displays a standard message box with the specified %text and %title,
   and the specified %buttons.
   For example,
   | UI.Message "Choose a color", "Question", "Red Green Blue"
   displays a message box with three buttons: "Red", "Green", and "Blue".

   The buttons can be activated by hitting their first letter.
   In addition, the <kbd>Enter</kbd> key activates the first button, <kbd>ESC</kbd> activates the last one.

   This command returns the index of the pressed button in {UI.Result}.
   For example, if the user chose "Red" above, {UI.Result} will have the value 1 afterwards.

   In text mode, displays text and heading, and a list of first letters
   of the buttons, and waits for a matching keystroke.

   The last two parameters are optional and default to "Message" and "OK".

   This command differs from {MessageBox} in that it modifies {UI.Result}, and waits for a keystroke in text mode.

   @diff PCC 1.x allows up to 10 buttons; PCC2 has no such limit
   (but you are adviced to keep the number of buttons and the length of the texts short anyway).

   @see MessageBox, UI.Input
   @since PCC 1.0.9, PCC2 1.99.9 */
static void
IFUIMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    using std::tolower;

    args.checkArgumentCount(1, 3);

    // Read arguments
    Ptr<RichText> text;
    string_t title   = _("Message");
    string_t buttons = _("OK");
    if (!checkRichArg(text, args.getNext()))
        return;
    checkStringArg(title, args.getNext());
    checkStringArg(buttons, args.getNext());

    buttons = strTrim(buttons);

    // Prepare dialog
    UIMessageBox msgbox(*text, title);
    int id = 0;
    do {
        string_t b = strFirst(buttons, " ");
        if (!b.empty()) {
            ++id;
            // Note: the PCC2 engine does not allow full UTF-8 support here because we have
            // an overlap between unicode codepoints and nonprinting characters. That is,
            // UI.Message "a","b",chr(273) will allow confirmation using SDLK_UP==273.
            // This can only be fixed by reworking key handling (-> c2ng).
            uint32_t key = charAtUtf8(b, 0);
            if (key < 256) {
                key = tolower(key);
            }
            msgbox.addButton(id, b, key);
        }
    } while (strRemove(buttons, " "));
    msgbox.addKey(1, SDLK_RETURN);
    msgbox.addKey(1, ' ');
    msgbox.addKey(id, SDLK_ESCAPE);

    // Do it
    if (id == 0) {
        exc.setVariable("UI.RESULT", 0);
    } else {
        flushUI();
        if (exc.checkForBreak())
            return;
        int result = msgbox.run();
        IntIntIntValue iv(result);
        exc.setVariable("UI.RESULT", &iv);
    }
}

/* @q UI.GotoChart x:Int, y:Int (Global Command)
   Go to starchart.
   This command activates the <a href="pcc2:starchart">starchart</a> at the specified position.
   If the coordinates are out of range, they are corrected.
   To switch to a the starcharts without affecting the current position, use
   | UI.GotoScreen 4

   @see UI.GotoScreen
   @since PCC 1.0.14, PCC2 1.99.10 */
static void
IFUIGotoChart(IntExecutionContext& exc, IntArgBlock& args)
{
    // Read arguments
    args.checkArgumentCount(2);
    int x, y;
    if (!checkIntArg(x, args.getNext(), 0, 10000) || !checkIntArg(y, args.getNext(), 0, 10000)) {
        return;
    }

    // Do it
    if (!postGoToChart(IntUserInterfaceBinding::get(), GPoint(x, y), exc.getCurrentObject())) {
        throw IntError::rangeError();
    }
    exc.suspendForUI();
}

/* @q UI.GotoScreen screen:Int, Optional id:Int (Global Command)
   Go to control screen.
   This command activates the specified screen.
   If the %id is specified and valid, shows that object.

   <table>
    <tr><th width="2">Id</th><th width="20" align="left">Screen</th></tr>
    <tr><td>0</td> <td><a href="pcc2:racescreen">Race screen</a>. You can not specify an %id here.</td></tr>
    <tr><td>1</td> <td><a href="pcc2:shipscreen">Ship screen</a>. %id is a ship Id.</td></tr>
    <tr><td>2</td> <td><a href="pcc2:planetscreen">Planet screen</a>. %id is a planet Id.</td></tr>
    <tr><td>3</td> <td><a href="pcc2:basescreen">Starbase screen</a>. %id is a starbase Id.</td></tr>
    <tr><td>4</td> <td><a href="pcc2:starchart">Starchart</a>. You can not specify an %id here. Also see {UI.GotoChart}.</td></tr>
    <tr><td>6</td> <td><a href="pcc2:historyscreen">Starship history</a>. %id is a ship Id.</td></tr>
    <tr><td>10</td><td><a href="pcc2:fleetscreen">Fleet screen</a>. %id is a fleet Id.</td></tr>
    <tr><td>11</td><td><a href="pcc2:shiptaskscreen">Ship auto task screen</a>. %id is a ship Id.</td></tr>
    <tr><td>12</td><td><a href="pcc2:planettaskscreen">Planet auto task screen</a>. %id is a planet Id.</td></tr>
    <tr><td>13</td><td><a href="pcc2:basetaskscreen">Starbase auto task screen</a>. %id is a starbase Id.</td></tr>
   </table>

   Note that this command will have immediate effect.
   It will suspend your script temporarily, switch to the new screen, and resume.

   @see UI.GotoChart
   @since PCC 1.0.14, PCC2 1.99.10 */
static void
IFUIGotoScreen(IntExecutionContext& exc, IntArgBlock& args)
{
    // Read arguments
    args.checkArgumentCount(1, 2);
    int32_t screen = 0;
    int32_t obj = 0;
    if (!checkIntArg(screen, args.getNext(), 0, 20))
        return;

    // If second argument is specified but empty, ignore command!
    if (args.getNumArgs() > 0 && !checkIntArg(obj, args.getNext(), 0, 10000))
        return;

    // Do it
    if (!postGoToScreen(IntUserInterfaceBinding::get(), screen, obj)) {
        throw IntError::rangeError();
    }
    exc.suspendForUI();
}

/* @q UI.PopupConsole (Global Command)
   Open the <a href="pcc2:console">console</a>.
   The script continues running there.
   If your script is doing interesting output to the console,
   you can call this function to ensure the user sees it, even if he bound it to a key.

   @since PCC 1.1.2, PCC2 1.99.10 */
static void
IFUIPopupConsole(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucPopupConsole));
    IntUserInterfaceBinding::get().postEvent(event);
    exc.suspendForUI();
}

/* @q UI.OverlayMessage msg:Str (Global Command)
   Display an overlay message.
   The message is shown centered on the screen, and automatically decays.
   Because it's not a window, the user doesn't have to explicitly confirm it.
   Use this for status updates from scripts that are not otherwise interactive.
   For example, this command is used to report changes of the current selection layer.
   @since PCC2 1.99.10 */
static void
IFUIOverlayMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    string_t msg;
    if (checkStringArg(msg, args.getNext()) && msg.size()) {
        flushUI();

        std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucOverlayMessage));
        event->args.pushNew(makeStringValue(msg));
        IntUserInterfaceBinding::get().postEvent(event);
        exc.suspendForUI();
    }
}

/* @q UI.ListShips x:Int, y:Int, Optional flags:Any, ok:Str, heading:Str (Global Command)
   List ships (visual scanner).

   Lists all ships at position %x,%y using the <a href="pcc2:listship">Visual Scan</a> window.
   The last three parameters are optional and modify behaviour details.

   The %flags parameter contains a list of flag letters:
   - "f": allow the user to choose foreign ships. If this is not specified,
     the "OK" button will be disabled for foreign ships. This flag implies "A".
   - "a": list all ships at the specified location. By default, only your ships are listed.
   - "e": when there is only one matching ship, return it and do not display the dialog at all.
   - "s": only show ships that we "safely" see, i.e. no guessed ships.
   - a ship Id to exclude. That ship will not be listed.

   The %ok string specifies the name of the "OK" button, it defaults to <tt>"OK"</tt>.
   Likewise, the %heading specifies the window title, it defaults to <tt>"List Ships"</tt>.

   The chosen ship Id (or EMPTY if the user canceled) is returned in {UI.Result}.
   If no ship matches, a dialog is displayed and EMPTY is returned.
   This command can't be used in text mode.

   For example, this command sequence sets a "Tow" mission:
   | UI.ListShips Loc.X, Loc.Y, "fae" & Id, "Choose", "Tow Ship"
   | If UI.Result Then SetMission 7, 0, UI.Result
   This command is equivalent to the <kbd>Ctrl-F1</kbd> key command (switch to ship):
   | UI.ListShips UI.X, UI.Y, "e" & Id
   | If UI.Result Then UI.GotoScreen 1, UI.Result

   @since PCC 1.1.1, PCC2 1.99.10
   @see UI.ChooseObject, UI.ListShipPrediction */
static void
IFUIListShips(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.ListShips x, y[, flags, ok, heading]
       Flags: F = allow selection of foreign ships; implies A
              A = list all ships
              E = do not display a dialog if there is only one ship
              S = only safe ships, no guessed ones */
    args.checkArgumentCount(2, 5);

    /* Read args */
    int32_t x, y;
    int32_t flags = 0, except = 0;
    string_t ok = _("OK");
    string_t heading = _("List Ships");

    if (!checkIntArg(x, args.getNext(), 0, 10000))
        return;
    if (!checkIntArg(y, args.getNext(), 0, 10000))
        return;
    checkFlagArg(flags, &except, args.getNext(), "FAES");
    checkStringArg(ok, args.getNext());
    checkStringArg(heading, args.getNext());

    /* Validate */
    if (except < 0 || except > NUM_SHIPS)
        throw IntError::rangeError();
    mustHaveTurn();

    /* Prepare */
    flushUI();
    if (exc.checkForBreak())
        return;

    int dlgFlags = 0;
    if (flags & 1)
        dlgFlags |= ls_IncludeForeignShips | ls_AllowForeignShips;
    if (flags & 2)
        dlgFlags |= ls_IncludeForeignShips;
    if (flags & 4)
        dlgFlags |= ls_EarlyExit;
    if (flags & 8)
        dlgFlags |= ls_SafeShipsOnly;

    /* Do it */
    GObjectReference result = doListShips(GPoint(x, y), dlgFlags, except, heading, ok, exc.getCurrentObject());
    if (result.isValid()) {
        IntIntIntValue iv(result->getId());
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.ListShipPrediction x:Int, y:Int, Optional sid:Int, ok:Str, heading:Str (Global Command)
   List ship prediction (visual scanner).

   Computes future positions of all (known) ships and lists all those that will be at %x,%y
   using the <a href="pcc2:listship">Visual Scan</a> window.
   When the %sid parameter is given and refers to a valid ship Id, uses that ship's predicted
   position instead of %x,%y.

   The last three parameters are optional and modify behaviour details.
   The %ok string specifies the name of the "OK" button, it defaults to <tt>"OK"</tt>.
   Likewise, the %heading specifies the window title, it defaults to <tt>"Ship Prediction"</tt>.

   The chosen ship Id (or EMPTY if the user canceled) is returned in {UI.Result}.
   If no ship matches, a dialog is displayed and EMPTY is returned.
   This command can't be used in text mode.

   @since PCC2 1.99.26
   @see UI.ChooseObject, UI.ListShips */
static void
IFUIListShipPrediction(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.ListShipPrediction x, y[, sid, title, okname] */
    args.checkArgumentCount(2, 5);

    /* Read args */
    int32_t x, y;
    int32_t sid = 0;
    string_t ok = _("OK");
    string_t heading = _("Ship Prediction");
    if (!checkIntArg(x, args.getNext(), 0, 10000)
        || !checkIntArg(y, args.getNext(), 0, 10000))
    {
        return;
    }
    checkIntArg(sid, args.getNext(), 0, NUM_SHIPS);
    checkStringArg(ok, args.getNext());
    checkStringArg(heading, args.getNext());

    /* Validate and prepare */
    mustHaveTurn();
    flushUI();
    if (exc.checkForBreak()) {
        return;
    }

    /* Do it */
    GObjectReference result = doListNextShips(GPoint(x, y), sid, heading, ok, exc.getCurrentObject());
    if (result.isValid()) {
        IntIntIntValue iv(result->getId());
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/** @q UI.ListFleets x:Int, y:Int, Optional flags:Any, ok:Str, heading:Str (Global Command)
    Choose a fleet.
    Lists all fleets at the specified %x,%y.
    If the user chooses one, returns their Id in {UI.Result}.
    If the user cancels the dialog using <kbd>ESC</kbd>, {UI.Result} is set to EMPTY.

    The %flags parameter modifies the function's behaviour:
    - "a": list all fleets, ignore %x,%y.
    - "e": do not display a dialog if there's just one fleet.

    @since PCC2 1.99.17 */
static void
IFUIListFleets(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.ListFleets x, y[, flags, ok, heading]
       Flags: A = list all fleets (not just those at X,Y)
              E = do not display a dialog if there is only one fleet
       "ok" is currently ignored, but present for consistency with UI.ListShips */
    args.checkArgumentCount(2, 5);

    /* Read args */
    int32_t x, y;
    int32_t flags = 0, except = 0;
    string_t ok = _("OK");
    string_t heading = _("List Fleets");

    if (!checkIntArg(x, args.getNext(), 0, 10000))
        return;
    if (!checkIntArg(y, args.getNext(), 0, 10000))
        return;
    checkFlagArg(flags, &except, args.getNext(), "AE");
    checkStringArg(ok, args.getNext());
    checkStringArg(heading, args.getNext());

    /* Validate */
    if (except < 0 || except > NUM_SHIPS)
        throw IntError::rangeError();
    mustHaveTurn();

    /* Prepare */
    flushUI();
    if (exc.checkForBreak())
        return;

    int dlgFlags = 0;
    if (flags & 1)
        dlgFlags |= WFleetList::fAllFleets;
    if (flags & 2)
        dlgFlags |= WFleetList::fQuickOut;

    /* Do it */
    int result = doFleetList(heading, getDisplayedTurn().getCurrentUniverse(), GPoint(x, y), dlgFlags, except);
    if (result) {
        IntIntIntValue iv(result);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

/* @q UI.Search Optional query:Str, flags:Any (Global Command)
   Search.

   When called with no parameters, just opens the <a href="pcc2:search">Search</a> dialog.
   When a search query is present, it is immmediately evaluated.
   The %query parameter is the search string, the %flags specify the kind of search:
   - "P": include planets in search.
   - "S": include ships in search.
   - "B": include starbases in search.
   - "U": include UFOs in search.
   - "O": include the other objects in search.
   - "M": include only my units in the search (PCC 2.0.2+).
   - "1": search for name or Id.
   - "2": search for expression which is true (default).
   - "3": search for expression which is false.
   - "4": search for location.
   Briefly, letters correspond to the checklist in the top-left of the search window,
   digits correspond to the selection list in the top-right.
   You can specify any number of letters but only one digit.
   By default, all objects are searched for an expression which is true.

   @since PCC2 1.99.10, PCC 1.1.2 */
static void
IFUISearch(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Search [text, flags]
       flags: PSBUOM = planets, ships, bases, ufos, others, mine
              1234  = name, true, false, location */
    bool immediate = false;
    args.checkArgumentCount(0, 2);
    if (args.getNumArgs() > 0) {
        /* Fetch text */
        immediate = true;
        if (!checkStringArg(search_query.query, args.getNext()))
            return;
    }
    if (args.getNumArgs() > 0) {
        /* Fetch flags */
        int32_t kind = 1;
        int32_t objs = 0;
        CTASSERT(TUserSearchQuery::SearchShips == 1);
        CTASSERT(TUserSearchQuery::SearchPlanets == 2);
        CTASSERT(TUserSearchQuery::SearchBases == 4);
        CTASSERT(TUserSearchQuery::SearchUfos == 8);
        CTASSERT(TUserSearchQuery::SearchOthers == 16);
        CTASSERT(TUserSearchQuery::SearchMineOnly == 32);
        if (!checkFlagArg(objs, &kind, args.getNext(), "SPBUOM"))
            return;

        switch (kind) {
         case 1:  search_query.search_kind = TUserSearchQuery::MatchName;     break;
         case 2:  search_query.search_kind = TUserSearchQuery::MatchTrue;     break;
         case 3:  search_query.search_kind = TUserSearchQuery::MatchFalse;    break;
         case 4:  search_query.search_kind = TUserSearchQuery::MatchLocation; break;
         default: throw IntError::rangeError();
        }

        search_query.search_obj = objs;
    }
    mustHaveTurn();

    /* Do it */
    flushUI();
    if (exc.checkForBreak())
        return;

    doSearch(immediate, exc.getCurrentObject());
    exc.suspendForUI();
}

/* @q UI.SelectionManager (Global Command)
   Open <a href="pcc2:selectionmgr">selection manager</a>.
   @since PCC2 1.99.10 */
static void
IFUISelectionManager(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();
    doSelectionManager(exc.getCurrentObject());
}

/* @q UI.PlanetInfo pid:Int (Global Command)
   Open <a href="pcc2:envscreen">planet information</a> for planet %pid.
   @since PCC2 1.99.10 */
static void
IFUIPlanetInfo(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t pid;
    args.checkArgumentCount(1);
    if (!checkIntArg(pid, args.getNext(), 1, NUM_PLANETS))
        return;
    mustHaveTurn();
    flushUI();
    doPlanetScan(pid);
    // exc.suspendForUI();
}

/* @q UI.KeymapInfo (Global Command)
   Open <a href="pcc2:keymap">keymap debugger</a> for current keymap.
   @since PCC2 1.99.10 */
static void
IFUIKeymapInfo(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();

    std::auto_ptr<IntValue> val(IntUserInterfaceBinding::get().getProperty(iuiKeymap));
    if (IntKeymapValue* kv = dynamic_cast<IntKeymapValue*>(val.get())) {
        doKeymapDebugger(kv->getKeymap());
    }
}

/* @q UI.Help page:Str (Global Command)
   Open help screen.
   The help page name is passed as a parameter.

   - PCC2: Help pages names are strings.
     For example, <tt>UI.Help "int:name:ui.help"</tt> displays this help page.
   - PCC 1.x: Help pages are identified by numbers.

   @diff In PCC 1.x, it is a script error if the page does not exist.
   PCC2 silently displays an error page.

   @since PCC2 1.99.15, PCC 1.0.15 */
static void
IFUIHelp(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    string_t page;
    args.checkArgumentCount(1);
    if (!checkStringArg(page, args.getNext()))
        return;
    flushUI();
    doHelp(page);
}

/* @q System.ExitRace (Global Command)
   Leave current race.
   Saves the game and returns to the <a href="pcc2:gamesel">game selection</a> menu.
   @since PCC2 1.99.10 */
static void
IFSystemExitRace(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucExitRace));
    IntUserInterfaceBinding::get().postEvent(event);
    exc.suspendForUI();
}

/* @q System.ExitClient (Global Command)
   Leave PCC2.
   Saves the game and closes the program.
   This command will also terminate the current process (as if the {End} command had been used).
   @since PCC2 1.99.26 */
static void
IFSystemExitClient(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucExitClient));
    IntUserInterfaceBinding::get().postEvent(event);
    exc.setState(IntExecutionContext::psTerminated);
}

/* @q LoadResource name:Str (Global Command)
   Load a resource.
   You can specify all items you can also specify in <tt>cc-res.cfg</tt>.
   - name of a *.res file (optionally prefixed by "res:")
   - "wp:" followed by the name of a Winplan "BMP" directory
   - "wpvcr:" followed by the name of a Winplan "WPVCR.DLL" file
   - "dir:" followed by a directory name
   File and directory names should be absolute.

   If the given name does not refer to a valid resource item,
   a console message will be printed, but the command will not fail.

   Because PCC2 caches loaded resource elements,
   you should call this command as early as possible (in <tt>pcc2init.q</tt>, usually).

   For loading *.res files, also see the chapter on <a href="int:statement:plugins">plugins</a>;
   loading a *.res file using the <tt>ResourceFile</tt> plugin directive will also work in PlayVCR
   which has no script interpreter.

   @since PCC 1.0.19, PCC2 1.99.25 */
static void
IFLoadResource(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1);

    // Fetch argument
    string_t arg;
    if (!checkStringArg(arg, args.getNext()))
        return;

    // Check context.
    // We want "LoadResource" statements from a plugin's "Exec" to be registered
    // with the same origin as resources loaded directly.
    loadResource(ResManager::getInstance(), arg, getPluginContext(exc));
}

/* @q LoadHelpFile name:Str (Global Command)
   Load a help file.
   PCC2 help files are files in a custom XML format.
   Help files can be added using plugins (HelpFile= directive), or using this command.

   This command does not verify that the file actually exists;
   if the given name does not refer to a valid help file,
   a console message will be printed, but the command will not fail.

   @since PCC2 2.0.5 */
static void
IFLoadHelpFile(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1);

    // Fetch argument
    string_t arg;
    if (!checkStringArg(arg, args.getNext()))
        return;

    HelpIndex::instance.addFile(arg, getPluginContext(exc));
}

/** Internal global command "CC$CSBrowse". */
static void
IFCCCSBrowse(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* CC$CSBrowse <forwardFlag>, <onlyMarkedFlag> */
    args.checkArgumentCount(2);

    WControlScreen* cs = dynamic_cast<WControlScreen*>(&IntUserInterfaceBinding::get());
    if (!cs)
        throw IntError::contextError();

    bool forward, marked;
    if (checkBoolArg(forward, args.getNext()) && checkBoolArg(marked, args.getNext()))
        cs->doBrowse(forward, marked);
}

static void
IFCCChangePE(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isPlayable(GObject::Playable)) {
        throw IntError::contextError();
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // Ship PE change
        StringList d;
        d.add(0, _("0 - none"));
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            d.add(i, format("%X - %s", i, player_racenames.getShortName(i)));

        UIStandardListbox lb(0, -1, 240, d);
        lb.scrollToKey(sh->getPrimaryEnemy());
        if (lb.doStandardListbox(_("Change Primary Enemy"), "pcc2:shipscreen"))
            sh->setPrimaryEnemy(mp16_t(lb.getKey()));
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeMission(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isPlayable(GObject::Playable)) {
        throw IntError::contextError();
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        doShipMissionChange(*sh);
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        if (pl->hasBase()) {
            doBaseChangeMission(*pl);
        } else {
            throw IntError::contextError();
        }
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildStructures(IntExecutionContext& exc, IntArgBlock& args)
{
    int32_t screen = 0;
    args.checkArgumentCount(1);
    checkIntArg(screen, args.getNext(), 0, 2);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isPlayable(GObject::Playable)) {
        throw IntError::contextError();
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        doPlanetBuildScreen(*pl, screen);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeTech(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isPlayable(GObject::Playable) && pl->hasBase()) {
        doTechUpgradeDialog(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeSpeed(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (sh != 0 && sh->isPlayable(GObject::Playable)) {
        if (sh->isFleetMember()) {
            throw GError(GError::eFleet, GError::eFleet);
        } else {
            WShipSpeedSelector(*sh).doStandardDialog();
        }
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeWaypoint(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (sh != 0 && sh->isPlayable(GObject::Playable)) {
        if (sh->isFleetMember()) {
            throw GError(GError::eFleet, GError::eFleet);
        } else {
            doShipNavigationChart(getDisplayedTurn().getCurrentUniverse(), *sh);
        }
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeTaxes(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isPlayable(GObject::Playable)) {
        doTaxation(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCSellSupplies(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isPlayable(GObject::Playable)) {
        WSellSuppliesDialog::doDialog(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCSettings(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    doPreferencesDialog();
}

static void
IFCCBuySupplies(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isPlayable(GObject::Playable)) {
        WUndoSellSuppliesDialog::doDialog(getDisplayedTurn(), pl->getId());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCGive(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isPlayable(GObject::Playable)) {
        throw IntError::contextError();
    } else if (!host.isPHost() || !config.CPEnableGive()) {
        // ignore
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        // give planet
        doGivePlanet(*pl, getDisplayedTurn().getCommands(pl->getOwner()));
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // give ship
        doGiveShip(*sh, getDisplayedTurn().getCommands(sh->getOwner()));
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCShipSpec(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (sh != 0 && sh->getShipKind() != GShip::NoShip) {
        GShipQuery q;
        q.initForExistingShip(getDisplayedTurn().getCurrentUniverse(), sh->getId());
        WSpecView sv;
        sv.setQuery(q);
        sv.doStandardDialog();
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCShipCostCalc(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();
    if (exc.checkForBreak())
        return;
    doShipCostCalculator(dynamic_cast<GPlanet*>(exc.getCurrentObject()));
}

static void
IFCCBuildAmmo(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isPlayable(GObject::Playable) && pl->hasBase()) {
        doBaseBuildAmmo(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildShip(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isPlayable(GObject::Playable) && pl->hasBase()) {
        std::vector<int> hulls;
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
            if (int j = getTruehull(pl->getOwner(), i))
                hulls.push_back(j);
        WBaseShipBuildDialog bs(*pl, hulls);
        ui_root->centerWidget(bs);
        ui_root->add(bs);
        ui_root->run();
        ui_root->close(bs);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCAddToSim(IntExecutionContext& exc, IntArgBlock& args)
{
    bool ask = true;
    args.checkArgumentCount(0, 1);
    checkBoolArg(ask, args.getNext());
    if (GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        if (pl->isVisible()) {
            doPlanetAddToSim(*pl);
        } else {
            throw GError(GError::eNotOwner, _("Nothing known about this planet"));
        }
    } else if (GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject())) {
        flushUI();
        doShipAddToSim(*sh, ask);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildBase(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    if (GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        flushUI();
        doPlanetBuildBase(getDisplayedTurn().getCurrentUniverse(), *pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCTransferPlanet(IntExecutionContext& exc, IntArgBlock& args)
{
    bool unload;
    args.checkArgumentCount(1);
    if (!checkBoolArg(unload, args.getNext()))
        return;
    if (GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        mustBePlayed(*pl);
        flushUI();
        doPlanetTransfer(getDisplayedTurn().getCurrentUniverse(), *pl, unload);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCTransferShip(IntExecutionContext& exc, IntArgBlock& args)
{
    // Args:
    //   int: kind of transfer (0=ask, 1=ship, 2=planet)
    //   int: target (planet 0: unload)
    int32_t mode, target;
    args.checkArgumentCount(2);
    if (!checkIntArg(mode, args.getNext(), 0, 2) || !checkIntArg(target, args.getNext(), 0, INT_MAX))
        return;

    // Must be our ship
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh)
        throw IntError::contextError();
    mustBePlayed(*sh);

    // Parse mode/target
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    if (mode == 0) {
        doShipCargoTransfer(univ, *sh);
        return;
    }

    // Parse and validate mode/target
    if (mode == 1) {
        // Ship
        if (!univ.ty_any_ships.isValidIndex(target))
            throw GError(GError::eRange, _("This ship does not exist"));
        if (univ.getShip(target).getPos() != sh->getPos())
            throw GError(GError::ePos, _("Different position"));
        if (target == sh->getId())
            throw GError(GError::ePerm, GError::ePerm);
        doShipTransferFor(*sh, GObjectReference(current_anyship_type, target));
    } else {
        // Planet
        if (sh->getOrbitPlanetId() != target)
            throw GError(GError::ePos, _("Different position"));
        if (target == 0) {
            // Jettison
            doShipTransferFor(*sh, GObjectReference());
        } else {
            // Planet
            doShipTransferFor(*sh, GObjectReference(current_anyplanet_type, target));
        }
    }
}

static void
IFCCTransferMulti(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0, 1);

    // Fleet flag
    bool flag = false;
    checkBoolArg(flag, args.getNext());

    // Must be our ship
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh)
        throw IntError::contextError();
    mustBePlayed(*sh);
    if (flag && sh->getFleetNumber() == 0)
        throw GError(GError::eRange, _("Ship is not member of a fleet."));

    // Do it
    flushUI();
    doMultiTransfer(sh->getId(), flag);
}

static void
IFCCTransferUnload(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh)
        throw IntError::contextError();
    mustBePlayed(*sh);
    if (sh->getOrbitPlanetId() == 0)
        throw GError(GError::ePos, _("Ship is not orbiting a planet."));

    flushUI();
    doShipTransferFor(*sh, GObjectReference(current_anyplanet_type, sh->getOrbitPlanetId()));
}

static void
IFCCRemoteControl(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh)
        throw IntError::contextError();
    mustBePlayed(*sh);
    if (!host.isPHost() || !config.CPEnableRemote())
        throw GError(GError::ePerm, _("Feature not available."));

    flushUI();
    doRemoteControl(*sh, getDisplayedTurn().getCommands(sh->getOwner()));
}

static void
IFCCPopScreenHistory(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();

    std::auto_ptr<IntValue> reg(IntUserInterfaceBinding::get().getProperty(iuiScreenRegistered));
    if (popHistoryScreenId(getBoolValue(reg.get()))) {
        exc.suspendForUI();
    }
}

static void
IFCCListScreenHistory(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();

    std::auto_ptr<IntValue> reg(IntUserInterfaceBinding::get().getProperty(iuiScreenRegistered));
    if (doScreenHistoryDialog(getBoolValue(reg.get()))) {
        exc.suspendForUI();
    }
}

static void
IFCCViewInbox(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();
    doInboxScreen();
    exc.suspendForUI();
}

static void
IFCCExplainPrediction(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();

    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh)
        throw IntError::contextError();
    mustBePlayed(*sh);

    explainPrediction(*sh, getDisplayedTurn().getCurrentUniverse());
}


void
doUnitMessageScreen(const std::vector<int>& msgs, string_t title)
{
    if (!msgs.empty()) {
        GInbox& inbox = getDisplayedTurn().getInbox();
        GProxyMailbox proxy(inbox, msgs);
        proxy.importOuterCursor();

        WSimpleMailboxWindow win(title, proxy);
        ui_root->centerWidget(win);
        win.runModal(win);
    }
}

static void
IFCCViewMessages(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();

    GObject* obj = exc.getCurrentObject();
    if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        doUnitMessageScreen(sh->getAssociatedMessages(), sh->getName(GObject::LongName));
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        doUnitMessageScreen(pl->getAssociatedMessages(), pl->getName(GObject::LongName));
    } else {
        throw IntError::contextError();
    }

    exc.suspendForUI();
}

static void
IFCCViewUtilData(IntExecutionContext& exc, IntArgBlock& args)
{
    // Quick & Dirty.
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();

    GUtilMessageParser p;
    p.loadDefinition(*spec_file_dir->openFile("utildata.cc", Stream::C_READ));

    GSimpleMailbox mbx;
    p.loadFile(*spec_file_dir->openFile(format("util%d.dat", getPlayerId()), Stream::C_READ), mbx);

    if (mbx.getCount() != 0) {
        GInbox& inbox = getDisplayedTurn().getInbox();
        WSimpleMailboxWindow win(_("Messages (util.dat)"), mbx);
        ui_root->centerWidget(win);
        win.runModal(win);
        exc.suspendForUI();
    }
}

static void
IFCCSendMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();
    doSendOrReviseMessage();
    exc.suspendForUI();
}

/* @q Chart.SetView name:Str (Global Command)
   Set current view in starchart.
   This determines the visible panels and active keymaps.
   @since PCC2 1.99.10 */
static void
IFChartSetView(IntExecutionContext& exc, IntArgBlock& args)
{
    string_t name;
    args.checkArgumentCount(1);
    if (checkStringArg(name, args.getNext())) {
        // Check for starchart
        int32_t screenId;
        std::auto_ptr<IntValue> screenVal(IntUserInterfaceBinding::get().getProperty(iuiScreenNumber));
        if (!checkIntArg(screenId, screenVal.get()) || screenId != 4) {
            throw IntError::contextError();
        }
        screenVal.reset();

        // Do it
        std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucChartSetView));
        event->args.pushNew(makeStringValue(strUCase(name)));
        IntUserInterfaceBinding::get().postEvent(event);
        exc.suspendForUI();
    }
}

static void
IFCCCargoHistory(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    if (GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject())) {
        flushUI();
        showCargoHistory(*sh);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCTowFleetMember(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);

    // Must be ship
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (!sh) {
        throw IntError::contextError();
    }

    // Must be part of a fleet
    if (sh->getFleetNumber() == 0) {
        throw GError(GError::eRange, _("Ship is not member of a fleet."));
    }

    // We will now always do something on the UI, so flush it.
    flushUI();

    // Check whether this ship can tow
    if (const GMission* msn = GMissionList::getInstance().getMissionByNumber(GMission::msn_Tow, GPlayerSet(sh->getRealOwner()))) {
        if (!msn->worksOn(*sh)) {
            messageBox(_("This ship cannot tow."),
                       _("Tow"));
            return;
        }
    }

    // Check intercept
    if (isMissionLocked(*sh, mlOverrideLocks)) {
        messageBox(_("This ship is member of a fleet which is on a intercept course. "
                     "You cannot tow and intercept at the same time."),
                   _("Tow"));
        return;
    }

    // Build list of possible targets
    static const int cancelMarker = 9999;
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GObjectList list;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& target = univ.getShip(i);
        if (i != sh->getId()
            && target.getFleetNumber() == sh->getFleetNumber()
            && target.getPos() == sh->getPos())
        {
            list.addObject(current_anyship_type, i);
        }
    }
    if (sh->getMission() == GMission::msn_Tow) {
        list.addNonObject(cancelMarker, _("Cancel tow"));
    }

    // Anything?
    if (list.size() == 0) {
        messageBox(_("There is no other fleet member at this position."),
                   _("Tow"));
        return;
    }

    // Ask user
    WObjectList widget(0, list.size() < 5 ? 5 : list.size() > 20 ? 20 : list.size(),
                       30 * font_ems[FONT_NORMAL], list);
    GObjectReference result = widget.doStandardDialog2(_("Tow"), "pcc2:fleetscreen");
    if (result.isValid()) {
        if (result->getId() == cancelMarker) {
            // Cancel tow
            setFleetMission(univ, sh->getId(), 0, 0, 0);
        } else {
            // Set tow
            if (!univ.isShipTowed(sh->getId())
                || UIMessageBox(_("This ship is already being towed. Tow chains will not work.\n"
                                  "Continue anyway?"),
                                _("Set Mission")).doYesNoDialog())
            {
                setFleetMission(univ, sh->getId(), GMission::msn_Tow, 0, result->getId());
            }
        }
    }
}

/* @q CC$UseKeymap keymap:Str, prefix:Int (Internal)
   This is the backend to {UseKeymap}.
   It causes the next keystroke to be processed using the %keymap,
   and gives it %prefix as the prefix argument ({UI.Prefix}).
   @since PCC2 1.99.22 */
static void
IFCCUseKeymap(IntExecutionContext& exc, IntArgBlock& args)
{
    // Read arguments
    args.checkArgumentCount(2);

    // Build and submit message
    std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucUseKeymap));
    event->args.push(args.getNext());
    event->args.push(args.getNext());
    IntUserInterfaceBinding::get().postEvent(event);
    exc.suspendForUI();
}

/* @q CC$Reset x:Int, y:Int (Internal)
   Reset location dialog. */
static void
IFCCReset(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Fetch location
    int32_t x, y;
    if (!checkIntArg(x, args.getNext()) || !checkIntArg(y, args.getNext())) {
        return;
    }

    // Do we have a turn?
    mustHaveTurn();

    // Do it
    GReset reset;
    reset.addLocation(getCurrentTurn(), GPoint(x, y));
    if (doResetDialog(reset)) {
        reset.commit();
    }
}

// @since PCC2 2.40.8, 2.0.9
static void
IFCCManageBuildQueue(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    mustHaveTurn();
    flushUI();

    // Focus on planet if possible (but don't fail if not)
    int planetId = 0;
    if (const GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject())) {
        planetId = pl->getId();
    }

    doBuildQueueDialog(getDisplayedTurn(), planetId);
}

static bool
checkBreak()
{
    bool result = checkBreakEvent();
    if (result) {
        result = UIMessageBox(_("PCC2 is currently executing a script. "
                                "Do you want to abort that script?"),
                              _("Script Interpreter")).doYesNoDialog();
        ui_root->performDeferredRedraws();
    }
    return result;
}

static void
initInterpreterGuiInterface()
{
    // Values
    /* @q System.GUI:Bool (Global Property)
       Graphical interface flag.
       True if PCC is running with graphical interface, False if it is running in console mode. */
    defineGlobalValue("SYSTEM.GUI", makeBoolValue(1));
    defineGlobalValue("RADD",           new IntSimpleIndexableValue(0, IFRAdd,              0));
    defineGlobalValue("RLEN",           new IntSimpleIndexableValue(0, IFRLen,              0));
    defineGlobalValue("RLINK",          new IntSimpleIndexableValue(0, IFRLink,             0));
    defineGlobalValue("RMID",           new IntSimpleIndexableValue(0, IFRMid,              0));
    defineGlobalValue("RSTRING",        new IntSimpleIndexableValue(0, IFRString,           0));
    defineGlobalValue("RSTYLE",         new IntSimpleIndexableValue(0, IFRStyle,            0));
    defineGlobalValue("RXML",           new IntSimpleIndexableValue(0, IFRXml,              0));
    defineGlobalValue("LISTBOX",        new IntSimpleIndexableValue(0, IFListbox,           0));

    // Procedures
    defineGlobalProcedure("CC$ADDTOSIM",           IFCCAddToSim);
    defineGlobalProcedure("CC$BUILDAMMO",          IFCCBuildAmmo);
    defineGlobalProcedure("CC$BUILDBASE",          IFCCBuildBase);
    defineGlobalProcedure("CC$BUILDSHIP",          IFCCBuildShip);
    defineGlobalProcedure("CC$BUILDSTRUCTURES",    IFCCBuildStructures);
    defineGlobalProcedure("CC$BUYSUPPLIES",        IFCCBuySupplies);
    defineGlobalProcedure("CC$CARGOHISTORY",       IFCCCargoHistory);
    defineGlobalProcedure("CC$CHANGEMISSION",      IFCCChangeMission);
    defineGlobalProcedure("CC$CHANGEPE",           IFCCChangePE);
    defineGlobalProcedure("CC$CHANGESPEED",        IFCCChangeSpeed);
    defineGlobalProcedure("CC$CHANGETAXES",        IFCCChangeTaxes);
    defineGlobalProcedure("CC$CHANGETECH",         IFCCChangeTech);
    defineGlobalProcedure("CC$CHANGEWAYPOINT",     IFCCChangeWaypoint);
    defineGlobalProcedure("CC$CSBROWSE",           IFCCCSBrowse);
    defineGlobalProcedure("CC$EXPLAINPREDICTION",  IFCCExplainPrediction);
    defineGlobalProcedure("CC$GIVE",               IFCCGive);
    defineGlobalProcedure("CC$LISTSCREENHISTORY",  IFCCListScreenHistory);
    defineGlobalProcedure("CC$MANAGEBUILDQUEUE",   IFCCManageBuildQueue);
    defineGlobalProcedure("CC$POPSCREENHISTORY",   IFCCPopScreenHistory);
    defineGlobalProcedure("CC$REMOTECONTROL",      IFCCRemoteControl);
    defineGlobalProcedure("CC$RESET",              IFCCReset);
    defineGlobalProcedure("CC$SELLSUPPLIES",       IFCCSellSupplies);
    defineGlobalProcedure("CC$SENDMESSAGE",        IFCCSendMessage);
    defineGlobalProcedure("CC$SETTINGS",           IFCCSettings);
    defineGlobalProcedure("CC$SHIPCOSTCALC",       IFCCShipCostCalc),
    defineGlobalProcedure("CC$SHIPSPEC",           IFCCShipSpec);
    defineGlobalProcedure("CC$TOWFLEETMEMBER",     IFCCTowFleetMember);
    defineGlobalProcedure("CC$TRANSFERMULTI",      IFCCTransferMulti);
    defineGlobalProcedure("CC$TRANSFERPLANET",     IFCCTransferPlanet);
    defineGlobalProcedure("CC$TRANSFERSHIP",       IFCCTransferShip);
    defineGlobalProcedure("CC$TRANSFERUNLOAD",     IFCCTransferUnload);
    defineGlobalProcedure("CC$USEKEYMAP",          IFCCUseKeymap);
    defineGlobalProcedure("CC$VIEWINBOX",          IFCCViewInbox);
    defineGlobalProcedure("CC$VIEWMESSAGES",       IFCCViewMessages);
    defineGlobalProcedure("CC$VIEWUTILDATA",       IFCCViewUtilData);
    defineGlobalProcedure("CHART.SETVIEW",         IFChartSetView);
    defineGlobalProcedure("LOADHELPFILE",          IFLoadHelpFile);
    defineGlobalProcedure("LOADRESOURCE",          IFLoadResource);
    defineGlobalProcedure("MESSAGEBOX",            IFMessageBox);
    defineGlobalProcedure("SYSTEM.EXITCLIENT",     IFSystemExitClient);
    defineGlobalProcedure("SYSTEM.EXITRACE",       IFSystemExitRace);
    defineGlobalProcedure("UI.CHOOSEOBJECT",       IFUIChooseObject);
    defineGlobalProcedure("UI.EDITALLIANCES",      IFUIEditAlliances);
    defineGlobalProcedure("UI.FILEWINDOW",         IFUIFileWindow);
    defineGlobalProcedure("UI.GOTOCHART",          IFUIGotoChart);
    defineGlobalProcedure("UI.GOTOSCREEN",         IFUIGotoScreen);
    defineGlobalProcedure("UI.HELP",               IFUIHelp);
    defineGlobalProcedure("UI.INPUT",              IFUIInput);
    defineGlobalProcedure("UI.INPUTFCODE",         IFUIInputFCode);
    defineGlobalProcedure("UI.INPUTNUMBER",        IFUIInputNumber);
    defineGlobalProcedure("UI.KEYMAPINFO",         IFUIKeymapInfo);
    defineGlobalProcedure("UI.LISTFLEETS",         IFUIListFleets);
    defineGlobalProcedure("UI.LISTSHIPPREDICTION", IFUIListShipPrediction);
    defineGlobalProcedure("UI.LISTSHIPS",          IFUIListShips);
    defineGlobalProcedure("UI.MESSAGE",            IFUIMessage);
    defineGlobalProcedure("UI.OVERLAYMESSAGE",     IFUIOverlayMessage);
    defineGlobalProcedure("UI.PLANETINFO",         IFUIPlanetInfo);
    defineGlobalProcedure("UI.POPUPCONSOLE",       IFUIPopupConsole);
    defineGlobalProcedure("UI.SEARCH",             IFUISearch);
    defineGlobalProcedure("UI.SELECTIONMANAGER",   IFUISelectionManager);
    defineGlobalProcedure("UI.UPDATE",             IFUIUpdate);

    IntExecutionContext::setBreakHandler(checkBreak);
}

// Constraints: after mod_intgame_common, after mod_int, with mod_base_gui
const ModuleInitializer mod_intgame_gui[] = {
    { OnStartup, initInterpreterGuiInterface },
    { 0, 0 }
};
