/**
  *  \file io/textfile.cc
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "io/textfile.h"
#include "io/streams.h"
#include "cpluslib/assert.h"
#include "cpluslib/types.h"
#include "cpluslib/utf8.h"
#include "arch/platform.h"
#include "util/nls.h"
#include "util/except.h"

/** Constructor.
    \param s Stream to read */
TextFile::TextFile(Stream& s)
    : stream(s),
      line_nr(0),
      buf_ptr(0),
      buf_fill(0),
      eof(false),
      mode(Neutral),
      sys_newline(true),
      utf8(false),
      cs()                      // default-initialisation gives effectively latin-1
{ }

/** Destructor. */
TextFile::~TextFile()
{ }

/** Refill buffer. */
void
TextFile::refillBuffer()
{
    ASSERT(buf_ptr >= buf_fill);
    ASSERT(mode == Reading);
    if (!eof) {
        buf_ptr = 0;
        buf_fill = stream.read(buffer, sizeof(buffer));
        if (buf_fill == 0)
            eof = true;
    }
}

/** Flush buffer. */
void
TextFile::flushBuffer()
{
    ASSERT(mode == Writing);
    stream.writeT(buffer, buf_fill);
    buf_fill = buf_ptr = 0;
}

/** Set mode. */
void
TextFile::setMode(Mode m)
{
    if (m != mode) {
        if (mode == Reading) {
            // We are reading, and have likely read too much.
            stream.seek(stream.getPos() - buf_fill + buf_ptr);
        }
        if (mode == Writing) {
            // Just flush the buffer. Make sure the reader
            // doesn't try BOM snooping.
            flushBuffer();
            if (line_nr == 0)
                line_nr = 1;
        }
        buf_fill = buf_ptr = 0;
        eof = false;
        if (m == Writing) {
            // We want to write. Check that this is possible.
            // (It is not needed to implement an equivalent
            // check for reading, because setMode(Reading)
            // will normally be followed by a real call to
            // read() which generates the error; for writing,
            // there can be an arbitrarily long time between
            // buffering and actual writing.)
            if (!stream.canDo(Stream::C_WRITE)) {
                mode = Neutral;
                throw FileProblemException(stream, _("Writing not supported"));
            }
        }
        mode = m;
    }
}

/** Read a line from the text file. Accepts all possible newline conventions.
    \param line [out] Line
    \return true if line read, false if EOF reached */
bool
TextFile::getLine(string_t& line)
{
    // BOM snooping
    static const int NBOM = 3;
    static const uint8_t bom[NBOM] = { 0xEF, 0xBB, 0xBF };
    int snoop;
    if (line_nr == 0 && !utf8)
        snoop = 0;
    else
        snoop = -1;

    // Read line
    line.clear();
    int ch;
    while ((ch = getByte()) >= 0 && (ch != 10 && ch != 13)) {
        // Accumulate line
        if (utf8) {
            line.append(1, char(ch));
        } else {
            appendUtf8(line, cs.getUnicode(ch));
        }

        // BOM snooping
        if (snoop >= 0 && ch == bom[snoop]) {
            ++snoop;
            if (snoop >= NBOM) {
                utf8 = true;
                line.clear();
                snoop = -1;
            }
        } else {
            snoop = -1;
        }
    }

    // If CRLF, read the LF
    if (ch == 13 && peekByte() == 10)
        getByte();

    // If the file ends in Ctrl-Z, remove that
    if (ch < 0 && line.size() && line[line.size()-1] == 26)
        line.erase(line.size()-1);

    if (ch >= 0 || line.size()) {
        ++line_nr;
        return true;
    } else {
        return false;
    }
}

/** Read a byte from the text file.
    \return character read (>= 0), or -1 on EOF */
int
TextFile::getByte()
{
    setMode(Reading);
    if (buf_ptr >= buf_fill)
        refillBuffer();
    return eof ? -1 : uint8(buffer[buf_ptr++]);
}

/** Preview a byte from the text file without reading it. The byte is
    not removed from the stream until you call getByte().
    \return character read (>= 0), or -1 on EOF */
int
TextFile::peekByte()
{
    setMode(Reading);
    if (buf_ptr >= buf_fill)
        refillBuffer();
    return eof ? -1 : uint8(buffer[buf_ptr]);
}

/** Read raw bytes from the stream. */
std::size_t
TextFile::readBytes(void* data, std::size_t count)
{
    /* Quick and simple */
    uint8_t* p = static_cast<uint8_t*>(data);
    std::size_t did = 0;
    int ch;
    while (did < count && (ch = getByte()) >= 0) {
        p[did++] = ch;
    }
    return did;
}

/** Write text to file. */
void
TextFile::write(const string_t& line)
{
    if (utf8) {
        writeBytes(line.data(), line.size());
    } else {
        string_t tmp(cs.convertFromUtf8(line));
        writeBytes(tmp.data(), tmp.size());
    }
}

/** Write text to file. */
void
TextFile::write(const char* data)
{
    write(string_t(data));
}

/** Write a newline. */
void
TextFile::writeLine()
{
    if (sys_newline) {
        writeBytes(Platform::LINE_SEPARATOR, std::strlen(Platform::LINE_SEPARATOR));
    } else {
        writeBytes("\n", 1);
    }
}

/** Write text to file, and start a new line. */
void
TextFile::writeLine(const string_t& line)
{
    write(line);
    writeLine();
}

/** Write raw data to file. */
void
TextFile::writeBytes(const void* data, std::size_t count)
{
    const char* p = static_cast<const char*>(data);
    setMode(Writing);
    while (count > 0) {
        if (buf_fill == 0 && count >= sizeof(buffer)) {
            // More than one buffer full: write directly
            stream.writeT(p, count);
            count = 0;
        } else {
            // Write through buffer
            std::size_t avail = sizeof(buffer) - buf_fill;
            std::size_t now = count > avail ? avail : count;
            std::memcpy(&buffer[buf_fill], p, now);
            p += now;
            count -= now;
            buf_fill += now;
            if (buf_fill >= sizeof(buffer))
                flushBuffer();
        }
    }
}

/** Make sure our cache is consistent with the disk file. */
void
TextFile::sync()
{
    setMode(Neutral);
}

/** Get file position. */
Stream::fpos_t
TextFile::getPos()
{
    Stream::fpos_t p = stream.getPos();
    if (mode == Reading) {
        p -= buf_fill;
        p += buf_ptr;
    }
    if (mode == Writing) {
        p += buf_fill;
    }
    return p;
}

/** Get file size. When writing, the pretended file position might be
    larger than the actual one on the stream. */
Stream::fpos_t
TextFile::getSize()
{
    Stream::fpos_t p = stream.getSize();
    if (mode == Writing) {
        Stream::fpos_t newPos = stream.getPos() + buf_fill;
        if (newPos > p)
            p = newPos;
    }
    return p;
}

/** Set file position. */
void
TextFile::seek(Stream::fpos_t pos)
{
    sync();
    eof = false;
    ASSERT(buf_fill == 0);
    stream.seek(pos);
}

/** Set single-byte character set. */
void
TextFile::setCharacterSet(CharacterSet cs)
{
    this->cs = cs;
    this->utf8 = false;
}

/** Set UTF-8. */
void
TextFile::setUtf8()
{
    this->utf8 = true;
}
