/**
  *  \file ui/ui.h
  *  \brief Simple User Interface Kernel
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_UI_H
#define PCC_V2_UI_UI_H

#include <vector>
#include "gfx/gfx.h"
#include "ui/event.h"
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/sigslot.h"
#include "cpluslib/holder.h"

class UILayoutManager;
class UISkin;

class UIBaseWidget : public Object {
    UIBaseWidget(const UIBaseWidget&);
    UIBaseWidget& operator=(const UIBaseWidget&);

 protected:
    GfxRect            extent;          /**< Position and size of widget. Absolute coordinates. */
    UIBaseWidget*      parent;          /**< Parent of this widget. */
    UIBaseWidget*      next_sibling;    /**< Next sibling. */
    UIBaseWidget*      prev_sibling;    /**< Previous sibling. */
    UIBaseWidget*      first_child;     /**< First child. */
    UIBaseWidget*      last_child;      /**< Last child. */
    UIBaseWidget*      focused_child;   /**< Focused child. */
    UIBaseWidget*      selected_child;  /**< Selected child. */
    UISkin*            skin;            /**< Skin. */

    int       status;                   /**< Status flags. These change during lifetime. */
    int       id;                       /**< Identifier. */

 public:
    struct LayoutInfo {
        /** Flags for layout (li_XXX). */
        int      flags;
        /** Minimum size. Layout manager should make sure widget does
            not get smaller than this. */
        GfxPoint min_size;
        /** Preferred size. When asked to compute the optimum size (pack()),
            all widgets come out with this size if possible. */
        GfxPoint pref_size;
        LayoutInfo();
    };
    enum {
        /** Widget does not want to be layouted. Layout managers will
            ignore it. */
        li_NoLayout = 1,
        /** Widget is flexible in horizontal direction. It might be enlarged when there is
            more space than needed. */
        li_FlexibleH = 2,
        /** Widget is flexible in vertical direction. It might be enlarged when there is
            more space than needed. */
        li_FlexibleV = 4,
        /** Widget is flexible in all directions. */
        li_Flexible  = li_FlexibleH | li_FlexibleV
    };
    enum {
        st_Exposed     = 1,     ///< Component is visible.
        st_Focused     = 2,     ///< Component has physical input focus. The user has actively selected it to work with.
        st_Selected    = 4,     ///< Component has logical input focus. The user is actually interacting with this component.
        st_Disabled    = 8,     ///< Component cannot be selected.
        st_IdleNotify  = 16,    ///< Component wants to receive `Idle' events. This flag only works for leaf widgets.
        st_Modal       = 32,    ///< Component is modal.
        st_WantRedraw  = 64,    ///< Component wishes to be redrawn.
        st_ChildRedraw = 128,   ///< Component has a child that wishes to be redrawn.

        st_Last        = 128    ///< Last status value. If a child widget needs own flags, those start at st_Last*2.
    };
    enum FocusState {
        NoFocus,                ///< This widget doesn't have focus.
        PrimaryFocus,           ///< This widget has the primary focus and is receiving input.
        BackgroundFocus         ///< This widget has focus, but its owner hasn't.
    };
    Object* layout_arg;         ///< Layout argument. Uninterpreted by UIBaseWidget. Can be interpreted by layout manager.

    /* Birth & Death */
    UIBaseWidget(int id = 0) throw();
    virtual ~UIBaseWidget();

    /* Look & Feel */
    virtual void          drawContent(GfxCanvas& canvas) = 0;                       // draw content of widget, end-users override this
    virtual void          drawEverything(GfxCanvas& canvas) = 0;                    // draw everything including children, bases override this
    virtual void          drawBelow(UIBaseWidget& child, GfxCanvas& canvas) = 0;    // draw everything below child, bases override this
    virtual GfxCanvas&    getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr) = 0; // get canvas to draw child on, bases override this

    void                  performDeferredDraw(GfxCanvas& canvas);

    virtual bool          handleEvent(const UIEvent& event, bool second_pass) = 0;   // can be overridden by whoever needs it, base provides default
    virtual bool          handleCommand(const UICommand& cmd, UIBaseWidget& sender); // can be overridden by whoever needs it
    virtual bool          isInside(GfxPoint pt);                                     // can be overridden by whoever needs it

    void                  setSkin(UISkin& skin, bool recursive = false);
    UISkin&               getSkin() { return *skin; }
    bool                  hasSkin() const { return skin; }

    /* Status Management */
    int                   getState() const { return status; }
    bool                  hasState(int what) const { return (status & what) != 0; }
    UIBaseWidget&         setState(int astat, bool enable);
    void                  setStateOf(int id, int astat, bool enable);
    virtual void          onStateChange(int bits, bool enabled);
    void                  updateIdleNotify(bool enable);

    void                  setFocusedWidget();
    void                  setFocusedWidget(UIBaseWidget& the_widget);
    void                  setFocusedChild(UIBaseWidget& the_widget);
    UIBaseWidget*         getFocusedChild() { return focused_child; }

    void                  setSelectedWidget();
    void                  setSelectedWidget(UIBaseWidget& the_widget);
    void                  clearSelectedWidget();

    void                  drawWidget(bool immediate);
    void                  drawWidget();
    SignalHandler*        observe(Signal0& sig);
    SignalHandler*        observe(Signal1<int>& sig);

    /* Administrivia */
    UIBaseWidget*         getParent() const { return parent; }
    UIBaseWidget*         getNextSibling() const { return next_sibling; }
    UIBaseWidget*         getPreviousSibling() const { return prev_sibling; }
    UIBaseWidget*         getFirstChild() const { return first_child; }
    UIBaseWidget*         getLastChild() const { return last_child; }
    void                  addChildWidget(UIBaseWidget& the_widget, UIBaseWidget* add_after);
    void                  removeChildWidget(UIBaseWidget& the_widget);
    virtual void          onAddRemoveChild(UIBaseWidget& the_widget, bool added) = 0; // overridden in base

    int                   getId() const { return id; }
    UIBaseWidget*         getWidgetById(int id);
    UIBaseWidget*         getFocusedWidget();
    FocusState            getFocusState() const;

    const GfxRect&        getExtent() const { return extent; }

    /* Size */
    virtual void          getLayoutInfo(LayoutInfo&) = 0;                          // overridden by everyone

    void                  setSize(int w, int h);
    void                  setOrigin(int x, int y);
    void                  setExtent(GfxRect newpos, bool by_layout = false);
    virtual void          onResize(GfxRect oldsize, bool by_layout) = 0;           // overridden in base
    virtual void          onChildResize(UIBaseWidget& child, GfxRect oldsize) = 0; // overridden in base

    /* Prefix Argument */
    static void           setPrefixArg(int);
    static int            consumePrefixArg();

    /* Main loops */
    virtual int           run();                 // overridden in root
    virtual void          stop(int arg);         // overridden in root

    /* Object */
    string_t              toString();

    Signal1<int>    sig_focus_changed;
};

class UIBaseComplexWidget : public UIBaseWidget {
 public:
    UIBaseComplexWidget(int id = 0) throw();
    ~UIBaseComplexWidget();

    virtual void          drawEverything(GfxCanvas& canvas);
    virtual void          drawBelow(UIBaseWidget& child, GfxCanvas& canvas);
    virtual GfxCanvas&    getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr);

    virtual bool          handleEvent(const UIEvent& event, bool second_pass);  // can be overridden in child

    virtual void          onAddRemoveChild(UIBaseWidget& the_widget, bool added);

    virtual void          onChildResize(UIBaseWidget& child, GfxRect oldsize);
};

/** Standard widget that can contain other widgets side-by-side, but has no layout for them. */
class UIWidget : public UIBaseComplexWidget {
 public:
    UIWidget(int id = 0) throw();
    ~UIWidget();

    virtual void          onResize(GfxRect oldsize, bool by_layout);

    void add(UIBaseWidget& the_widget, GfxRect size);
};

/** Standard widget that can contain other widgets and has layout. */
class UILayoutableWidget : public UIBaseComplexWidget {
    UILayoutManager* layout;
    void adjustLayoutSize(GfxPoint& pt);
 public:
    UILayoutableWidget(UILayoutManager& layout, int id = 0) throw();
    ~UILayoutableWidget();

    void                  setLayout(UILayoutManager& layout);

    virtual void          onResize(GfxRect oldsize, bool by_layout);
    virtual void          getLayoutInfo(LayoutInfo&);
    virtual void          adjustSize(GfxRect& theSize, bool to_inner) = 0;

    void add(UIBaseWidget& the_widget);
    void pack();
    void doLayout();
};

class UIGroup : public UILayoutableWidget {
 public:
    UIGroup(UILayoutManager& layout, int id = 0) throw();

    void adjustSize(GfxRect& theSize, bool to_inner);
    void drawContent(GfxCanvas& canvas);
};

class UIFixedWidget : public UIWidget {
 public:
    UIFixedWidget(int id = 0) throw();
    ~UIFixedWidget();

    void getLayoutInfo(LayoutInfo& info);
};

/// Spacer.
class UISpacer : public UIWidget {
 public:
    UISpacer() throw();
    ~UISpacer();
    void drawContent(GfxCanvas& canvas);
    void getLayoutInfo(LayoutInfo& info);
};


/********************************* Inline ********************************/

inline void
UILayoutableWidget::setLayout(UILayoutManager& layout)
{
    this->layout = &layout;
}

#endif
